# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'game_input_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_GameInputDialog(object):
    def setupUi(self, GameInputDialog):
        if not GameInputDialog.objectName():
            GameInputDialog.setObjectName(u"GameInputDialog")
        GameInputDialog.resize(407, 235)
        self.gridLayout = QGridLayout(GameInputDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.accept_button = QPushButton(GameInputDialog)
        self.accept_button.setObjectName(u"accept_button")

        self.gridLayout.addWidget(self.accept_button, 6, 0, 1, 1)

        self.auto_save_spoiler_check = QCheckBox(GameInputDialog)
        self.auto_save_spoiler_check.setObjectName(u"auto_save_spoiler_check")

        self.gridLayout.addWidget(self.auto_save_spoiler_check, 5, 0, 1, 2)

        self.output_file_button = QPushButton(GameInputDialog)
        self.output_file_button.setObjectName(u"output_file_button")

        self.gridLayout.addWidget(self.output_file_button, 4, 2, 1, 1)

        self.output_file_edit = QLineEdit(GameInputDialog)
        self.output_file_edit.setObjectName(u"output_file_edit")

        self.gridLayout.addWidget(self.output_file_edit, 4, 0, 1, 2)

        self.input_file_label = QLabel(GameInputDialog)
        self.input_file_label.setObjectName(u"input_file_label")
        self.input_file_label.setMaximumSize(QSize(16777215, 20))

        self.gridLayout.addWidget(self.input_file_label, 1, 0, 1, 2)

        self.description_label = QLabel(GameInputDialog)
        self.description_label.setObjectName(u"description_label")
        self.description_label.setWordWrap(True)

        self.gridLayout.addWidget(self.description_label, 0, 0, 1, 3)

        self.input_file_edit = QLineEdit(GameInputDialog)
        self.input_file_edit.setObjectName(u"input_file_edit")

        self.gridLayout.addWidget(self.input_file_edit, 2, 0, 1, 2)

        self.cancel_button = QPushButton(GameInputDialog)
        self.cancel_button.setObjectName(u"cancel_button")

        self.gridLayout.addWidget(self.cancel_button, 6, 2, 1, 1)

        self.output_file_label = QLabel(GameInputDialog)
        self.output_file_label.setObjectName(u"output_file_label")
        self.output_file_label.setMaximumSize(QSize(16777215, 20))

        self.gridLayout.addWidget(self.output_file_label, 3, 0, 1, 2)

        self.input_file_button = QPushButton(GameInputDialog)
        self.input_file_button.setObjectName(u"input_file_button")

        self.gridLayout.addWidget(self.input_file_button, 2, 2, 1, 1)


        self.retranslateUi(GameInputDialog)

        QMetaObject.connectSlotsByName(GameInputDialog)
    # setupUi

    def retranslateUi(self, GameInputDialog):
        GameInputDialog.setWindowTitle(QCoreApplication.translate("GameInputDialog", u"Game Patching", None))
        self.accept_button.setText(QCoreApplication.translate("GameInputDialog", u"Accept", None))
        self.auto_save_spoiler_check.setText(QCoreApplication.translate("GameInputDialog", u"Include a spoiler log on same directory", None))
        self.output_file_button.setText(QCoreApplication.translate("GameInputDialog", u"Select File", None))
        self.output_file_edit.setPlaceholderText(QCoreApplication.translate("GameInputDialog", u"Path where to place randomized game", None))
        self.input_file_label.setText(QCoreApplication.translate("GameInputDialog", u"Input File (Vanilla Gamecube ISO)", None))
        self.description_label.setText(QCoreApplication.translate("GameInputDialog", u"<html><head/><body><p>In order to create the randomized game, a ISO file of Metroid Prime 2: Echoes for the Nintendo Gamecube is necessary.</p><p>After using it once, a copy is kept by Randovania for later use.</p></body></html>", None))
        self.input_file_edit.setPlaceholderText(QCoreApplication.translate("GameInputDialog", u"Path to vanilla Gamecube ISO", None))
        self.cancel_button.setText(QCoreApplication.translate("GameInputDialog", u"Cancel", None))
        self.output_file_label.setText(QCoreApplication.translate("GameInputDialog", u"Output File", None))
        self.input_file_button.setText(QCoreApplication.translate("GameInputDialog", u"Select File", None))
    # retranslateUi




from pathlib import Path

from randovania.dol_patching.dol_file import DolFile
from randovania.dol_patching.dol_version import find_version_for_dol
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.games.prime import echoes_dol_versions, all_prime_dol_patches, echoes_dol_patches
from randovania.games.prime.all_prime_dol_patches import BasePrimeDolVersion
from randovania.games.prime.echoes_dol_patches import EchoesDolVersion
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences

ALL_VERSIONS_PATCHES = echoes_dol_versions.ALL_VERSIONS


def apply_patches(game_root: Path, game_specific: EchoesGameSpecific, user_preferences: EchoesUserPreferences,
                  default_items: dict):
    dol_file = DolFile(_get_dol_path(game_root))

    version = find_version_for_dol(dol_file, ALL_VERSIONS_PATCHES)
    if not isinstance(version, BasePrimeDolVersion):
        return

    dol_file.set_editable(True)
    with dol_file:
        all_prime_dol_patches.apply_remote_execution_patch(version.string_display, dol_file)
        all_prime_dol_patches.apply_energy_tank_capacity_patch(version.health_capacity, game_specific, dol_file)
        all_prime_dol_patches.apply_reverse_energy_tank_heal_patch(version.sda2_base, version.dangerous_energy_tank,
                                                                   game_specific.dangerous_energy_tank,
                                                                   version.game, dol_file)

        if isinstance(version, EchoesDolVersion):
            echoes_dol_patches.apply_game_options_patch(version.game_options_constructor_address,
                                                        user_preferences, dol_file)
            echoes_dol_patches.apply_beam_cost_patch(version.beam_cost_addresses, game_specific, dol_file)
            echoes_dol_patches.apply_safe_zone_heal_patch(version.safe_zone, version.sda2_base, game_specific, dol_file)
            echoes_dol_patches.apply_starting_visor_patch(version.starting_beam_visor, default_items, dol_file)


def _get_dol_path(game_root: Path) -> Path:
    return game_root.joinpath("sys/main.dol")



from typing import Optional, Tuple, Callable, FrozenSet

from randovania.game_description import data_reader
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.node import PickupNode, ResourceNode, EventNode, Node
from randovania.game_description.requirements import RequirementSet, RequirementList
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.resolver import debug, event_pickup
from randovania.resolver.bootstrap import logic_bootstrap
from randovania.resolver.event_pickup import EventPickupNode
from randovania.resolver.logic import Logic
from randovania.resolver.resolver_reach import ResolverReach
from randovania.resolver.state import State


def _simplify_requirement_list(self: RequirementList, state: State,
                               dangerous_resources: FrozenSet[ResourceInfo],
                               ) -> Optional[RequirementList]:
    items = []
    for item in self.values():
        if item.negate:
            return None

        if item.satisfied(state.resources, state.energy):
            continue

        if item.resource not in dangerous_resources:
            # An empty RequirementList is considered satisfied, so we don't have to add the trivial resource
            items.append(item)

    return RequirementList(items)


def _simplify_additional_requirement_set(requirements: RequirementSet,
                                         state: State,
                                         dangerous_resources: FrozenSet[ResourceInfo],
                                         ) -> RequirementSet:
    new_alternatives = [
        _simplify_requirement_list(alternative, state, dangerous_resources)
        for alternative in requirements.alternatives
    ]
    return RequirementSet(alternative
                          for alternative in new_alternatives

                          # RequirementList.simplify may return None
                          if alternative is not None)


def _should_check_if_action_is_safe(state: State,
                                    action: ResourceNode,
                                    dangerous_resources: FrozenSet[ResourceInfo],
                                    all_nodes: Tuple[Node, ...]) -> bool:
    """
    Determines if we should _check_ if the given action is safe that state
    :param state:
    :param action:
    :return:
    """
    if any(resource in dangerous_resources
           for resource in action.resource_gain_on_collect(state.patches, state.resources, all_nodes)):
        return False

    if isinstance(action, EventNode):
        return True

    if isinstance(action, EventPickupNode):
        pickup_node = action.pickup_node
    else:
        pickup_node = action

    if isinstance(pickup_node, PickupNode):
        target = state.patches.pickup_assignment.get(pickup_node.pickup_index)
        if target is not None and (target.pickup.item_category.is_major_category or target.pickup.item_category.is_key):
            return True

    return False


def _inner_advance_depth(state: State,
                         logic: Logic,
                         status_update: Callable[[str], None],
                         *,
                         reach: Optional[ResolverReach] = None,
                         ) -> Tuple[Optional[State], bool]:
    """

    :param state:
    :param logic:
    :param status_update:
    :param reach: A precalculated reach for the given state
    :return:
    """

    if logic.game.victory_condition.satisfied(state.resources, state.energy):
        return state, True

    if reach is None:
        reach = ResolverReach.calculate_reach(logic, state)

    debug.log_new_advance(state, reach)
    status_update("Resolving... {} total resources".format(len(state.resources)))

    for action, energy in reach.possible_actions(state):
        if _should_check_if_action_is_safe(state, action, logic.game.dangerous_resources,
                                           logic.game.world_list.all_nodes):

            potential_state = state.act_on_node(action, path=reach.path_to_node[action], new_energy=energy)
            potential_reach = ResolverReach.calculate_reach(logic, potential_state)

            # If we can go back to where we were, it's a simple safe node
            if state.node in potential_reach.nodes:
                new_result = _inner_advance_depth(state=potential_state,
                                                  logic=logic,
                                                  status_update=status_update,
                                                  reach=potential_reach)

                if not new_result[1]:
                    debug.log_rollback(state, True, True)

                # If a safe node was a dead end, we're certainly a dead end as well
                return new_result

    debug.log_checking_satisfiable_actions()
    has_action = False
    for action, energy in reach.satisfiable_actions(state, logic.game.victory_condition):
        new_result = _inner_advance_depth(
            state=state.act_on_node(action, path=reach.path_to_node[action], new_energy=energy),
            logic=logic,
            status_update=status_update)

        # We got a positive result. Send it back up
        if new_result[0] is not None:
            return new_result
        else:
            has_action = True

    debug.log_rollback(state, has_action, False)
    additional_requirements = reach.satisfiable_as_requirement_set

    if has_action:
        additional = set()
        for resource_node in reach.collectable_resource_nodes(state):
            additional |= logic.get_additional_requirements(resource_node).alternatives

        additional_requirements = additional_requirements.union(RequirementSet(additional))

    logic.additional_requirements[state.node] = _simplify_additional_requirement_set(additional_requirements,
                                                                                     state,
                                                                                     logic.game.dangerous_resources)
    return None, has_action


def advance_depth(state: State, logic: Logic, status_update: Callable[[str], None]) -> Optional[State]:
    return _inner_advance_depth(state, logic, status_update)[0]


def _quiet_print(s):
    pass


def resolve(configuration: EchoesConfiguration,
            patches: GamePatches,
            status_update: Optional[Callable[[str], None]] = None
            ) -> Optional[State]:
    if status_update is None:
        status_update = _quiet_print

    game = data_reader.decode_data(configuration.game_data)
    event_pickup.replace_with_event_pickups(game)

    new_game, starting_state = logic_bootstrap(configuration, game, patches)
    logic = Logic(new_game, configuration)
    starting_state.resources["add_self_as_requirement_to_resources"] = 1
    debug.log_resolve_start()

    return advance_depth(starting_state, logic, status_update)



_CURRENT_OPTIONS_FILE_VERSION = 12


def _convert_logic(layout_logic: str) -> str:
    if layout_logic == "no-glitches":
        return "no-tricks"
    return layout_logic


def _convert_v11(options: dict) -> dict:
    options.pop("layout_configuration", None)
    options.pop("patcher_configuration", None)

    return options


_CONVERTER_FOR_VERSION = {
    11: _convert_v11,
}


def get_persisted_options_from_data(persisted_data: dict) -> dict:
    version = persisted_data.get("version", 0)
    options = persisted_data.get("options")

    if not isinstance(options, dict):
        print("Data has no options.")
        return {}

    while version < _CURRENT_OPTIONS_FILE_VERSION:
        converter = _CONVERTER_FOR_VERSION.get(version)
        if converter is None:
            print("Converter not found for version '{}'".format(version))
            return {}

        options = converter(options)
        version += 1

    if version > _CURRENT_OPTIONS_FILE_VERSION:
        print("Options has an version from the future '{}'. Supported is only up to {}".format(
            version, _CURRENT_OPTIONS_FILE_VERSION))
        return {}

    return options


def serialized_data_for_options(data_to_persist: dict) -> dict:
    return {
        "version": _CURRENT_OPTIONS_FILE_VERSION,
        "options": data_to_persist
    }



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_hints.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetEchoesHints(object):
    def setupUi(self, PresetEchoesHints):
        if not PresetEchoesHints.objectName():
            PresetEchoesHints.setObjectName(u"PresetEchoesHints")
        PresetEchoesHints.resize(423, 259)
        self.centralWidget = QWidget(PresetEchoesHints)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.hint_layout = QVBoxLayout(self.centralWidget)
        self.hint_layout.setSpacing(6)
        self.hint_layout.setContentsMargins(11, 11, 11, 11)
        self.hint_layout.setObjectName(u"hint_layout")
        self.hint_layout.setContentsMargins(4, 8, 4, 0)
        self.hint_sky_temple_key_label = QLabel(self.centralWidget)
        self.hint_sky_temple_key_label.setObjectName(u"hint_sky_temple_key_label")
        self.hint_sky_temple_key_label.setWordWrap(True)

        self.hint_layout.addWidget(self.hint_sky_temple_key_label)

        self.hint_sky_temple_key_combo = QComboBox(self.centralWidget)
        self.hint_sky_temple_key_combo.addItem("")
        self.hint_sky_temple_key_combo.addItem("")
        self.hint_sky_temple_key_combo.addItem("")
        self.hint_sky_temple_key_combo.setObjectName(u"hint_sky_temple_key_combo")

        self.hint_layout.addWidget(self.hint_sky_temple_key_combo)

        PresetEchoesHints.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesHints)

        QMetaObject.connectSlotsByName(PresetEchoesHints)
    # setupUi

    def retranslateUi(self, PresetEchoesHints):
        PresetEchoesHints.setWindowTitle(QCoreApplication.translate("PresetEchoesHints", u"Hints", None))
        self.hint_sky_temple_key_label.setText(QCoreApplication.translate("PresetEchoesHints", u"<html><head/><body><p>This controls how precise the hints for Sky Temple Keys in Sky Temple Gateway are.</p><p><span style=\" font-weight:600;\">No hints</span>: The scans provide no useful information.</p><p><span style=\" font-weight:600;\">Show only the area name</span>: Each scan says the corresponding key is in &quot;Temple Grounds&quot;, &quot;Agon Wastes&quot;, etc.<br/>Aether and Dark Aether are distinguished; for example, &quot;Agon Wastes&quot; refers only to the light world.</p><p><span style=\" font-weight:600;\">Show area and room name</span>: Each scan says the corresponding key is in &quot;Great Temple - Transport A Access&quot;, &quot;Ing Hive - Hive Entrance&quot;, etc. For rooms with more than one item location, there's no way to distinguish which one of them that key is in.</p></body></html>", None))
        self.hint_sky_temple_key_combo.setItemText(0, QCoreApplication.translate("PresetEchoesHints", u"No hints", None))
        self.hint_sky_temple_key_combo.setItemText(1, QCoreApplication.translate("PresetEchoesHints", u"Show only the area name", None))
        self.hint_sky_temple_key_combo.setItemText(2, QCoreApplication.translate("PresetEchoesHints", u"Show area and room name", None))

    # retranslateUi




import dataclasses
from enum import Enum

from randovania.bitpacking.bitpacking import BitPackDataClass, BitPackEnum


class SkyTempleKeyHintMode(BitPackEnum, Enum):
    DISABLED = "disabled"
    HIDE_AREA = "hide-area"
    PRECISE = "precise"

    @classmethod
    def default(cls) -> "SkyTempleKeyHintMode":
        return cls.PRECISE


@dataclasses.dataclass(frozen=True)
class HintConfiguration(BitPackDataClass):
    item_hints: bool = True
    sky_temple_keys: SkyTempleKeyHintMode = SkyTempleKeyHintMode.default()

    @classmethod
    def default(cls) -> "HintConfiguration":
        return cls()

    @property
    def as_json(self) -> dict:
        return {
            "item_hints": self.item_hints,
            "sky_temple_keys": self.sky_temple_keys.value,
        }

    @classmethod
    def from_json(cls, value: dict) -> "HintConfiguration":
        params = {}

        if "sky_temple_keys" in value:
            params["sky_temple_keys"] = SkyTempleKeyHintMode(value["sky_temple_keys"])

        if "item_hints" in value:
            params["item_hints"] = value["item_hints"]

        return cls(**params)



from enum import Enum
from typing import Optional, Callable, Awaitable, List, NamedTuple, Dict, Tuple

from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry


class GameConnectionStatus(Enum):
    Disconnected = "disconnected"
    UnknownGame = "unknown-game"
    WrongGame = "wrong-game"
    WrongHash = "wrong-hash"
    TitleScreen = "title-screen"
    TrackerOnly = "tracker-only"
    InGame = "in-game"

    @property
    def pretty_text(self) -> str:
        return _pretty_connection_status[self]


_pretty_connection_status = {
    GameConnectionStatus.Disconnected: "Disconnected",
    GameConnectionStatus.UnknownGame: "Unknown game",
    GameConnectionStatus.WrongGame: "Wrong game",
    GameConnectionStatus.WrongHash: "Correct game, wrong seed hash",
    GameConnectionStatus.TitleScreen: "Title screen",
    GameConnectionStatus.TrackerOnly: "Tracker only",
    GameConnectionStatus.InGame: "In-game",
}


class InventoryItem(NamedTuple):
    amount: int
    capacity: int


class ConnectionBase:
    _location_collected_listener: Optional[Callable[[int], Awaitable[None]]] = None

    @property
    def current_status(self) -> GameConnectionStatus:
        raise NotImplementedError()

    def get_current_inventory(self) -> Dict[ItemResourceInfo, InventoryItem]:
        raise NotImplementedError()

    def set_permanent_pickups(self, pickups: List[Tuple[str, PickupEntry]]):
        raise NotImplementedError()

    def set_location_collected_listener(self, listener: Optional[Callable[[int], Awaitable[None]]]):
        self._location_collected_listener = listener

    async def _emit_location_collected(self, location: int):
        if self._location_collected_listener is not None:
            await self._location_collected_listener(location)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'node_details_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_NodeDetailsPopup(object):
    def setupUi(self, NodeDetailsPopup):
        if not NodeDetailsPopup.objectName():
            NodeDetailsPopup.setObjectName(u"NodeDetailsPopup")
        NodeDetailsPopup.resize(556, 475)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(NodeDetailsPopup.sizePolicy().hasHeightForWidth())
        NodeDetailsPopup.setSizePolicy(sizePolicy)
        NodeDetailsPopup.setMaximumSize(QSize(16777215, 16777215))
        self.main_layout = QVBoxLayout(NodeDetailsPopup)
        self.main_layout.setSpacing(6)
        self.main_layout.setContentsMargins(11, 11, 11, 11)
        self.main_layout.setObjectName(u"main_layout")
        self.main_layout.setContentsMargins(4, 4, 4, 4)
        self.name_layout = QHBoxLayout()
        self.name_layout.setSpacing(6)
        self.name_layout.setObjectName(u"name_layout")
        self.name_label = QLabel(NodeDetailsPopup)
        self.name_label.setObjectName(u"name_label")
        sizePolicy.setHeightForWidth(self.name_label.sizePolicy().hasHeightForWidth())
        self.name_label.setSizePolicy(sizePolicy)

        self.name_layout.addWidget(self.name_label)

        self.name_edit = QLineEdit(NodeDetailsPopup)
        self.name_edit.setObjectName(u"name_edit")

        self.name_layout.addWidget(self.name_edit)


        self.main_layout.addLayout(self.name_layout)

        self.heals_check = QCheckBox(NodeDetailsPopup)
        self.heals_check.setObjectName(u"heals_check")

        self.main_layout.addWidget(self.heals_check)

        self.location_group = QGroupBox(NodeDetailsPopup)
        self.location_group.setObjectName(u"location_group")
        self.location_group.setCheckable(True)
        self.location_group.setChecked(True)
        self.location_layout = QHBoxLayout(self.location_group)
        self.location_layout.setSpacing(6)
        self.location_layout.setContentsMargins(11, 11, 11, 11)
        self.location_layout.setObjectName(u"location_layout")
        self.location_layout.setContentsMargins(4, 2, 4, 2)
        self.location_x_label = QLabel(self.location_group)
        self.location_x_label.setObjectName(u"location_x_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.location_x_label.sizePolicy().hasHeightForWidth())
        self.location_x_label.setSizePolicy(sizePolicy1)
        self.location_x_label.setMinimumSize(QSize(20, 0))

        self.location_layout.addWidget(self.location_x_label)

        self.location_x_spin = QDoubleSpinBox(self.location_group)
        self.location_x_spin.setObjectName(u"location_x_spin")
        self.location_x_spin.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.location_x_spin.setMinimum(-999.990000000000009)
        self.location_x_spin.setMaximum(999.990000000000009)

        self.location_layout.addWidget(self.location_x_spin)

        self.location_y_label = QLabel(self.location_group)
        self.location_y_label.setObjectName(u"location_y_label")
        sizePolicy1.setHeightForWidth(self.location_y_label.sizePolicy().hasHeightForWidth())
        self.location_y_label.setSizePolicy(sizePolicy1)
        self.location_y_label.setMinimumSize(QSize(20, 0))

        self.location_layout.addWidget(self.location_y_label)

        self.location_y_spin = QDoubleSpinBox(self.location_group)
        self.location_y_spin.setObjectName(u"location_y_spin")
        self.location_y_spin.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.location_y_spin.setMinimum(-999.990000000000009)
        self.location_y_spin.setMaximum(999.990000000000009)

        self.location_layout.addWidget(self.location_y_spin)

        self.location_z_label = QLabel(self.location_group)
        self.location_z_label.setObjectName(u"location_z_label")
        sizePolicy1.setHeightForWidth(self.location_z_label.sizePolicy().hasHeightForWidth())
        self.location_z_label.setSizePolicy(sizePolicy1)
        self.location_z_label.setMinimumSize(QSize(20, 0))

        self.location_layout.addWidget(self.location_z_label)

        self.location_z_spin = QDoubleSpinBox(self.location_group)
        self.location_z_spin.setObjectName(u"location_z_spin")
        self.location_z_spin.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.location_z_spin.setMinimum(-999.990000000000009)
        self.location_z_spin.setMaximum(999.990000000000009)

        self.location_layout.addWidget(self.location_z_spin)


        self.main_layout.addWidget(self.location_group)

        self.node_type_combo = QComboBox(NodeDetailsPopup)
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.addItem("")
        self.node_type_combo.setObjectName(u"node_type_combo")

        self.main_layout.addWidget(self.node_type_combo)

        self.tab_widget = QTabWidget(NodeDetailsPopup)
        self.tab_widget.setObjectName(u"tab_widget")
        self.tab_generic = QWidget()
        self.tab_generic.setObjectName(u"tab_generic")
        self.generic_layout = QHBoxLayout(self.tab_generic)
        self.generic_layout.setSpacing(6)
        self.generic_layout.setContentsMargins(11, 11, 11, 11)
        self.generic_layout.setObjectName(u"generic_layout")
        self.label = QLabel(self.tab_generic)
        self.label.setObjectName(u"label")
        self.label.setAlignment(Qt.AlignCenter)

        self.generic_layout.addWidget(self.label)

        self.tab_widget.addTab(self.tab_generic, "")
        self.tab_dock = QWidget()
        self.tab_dock.setObjectName(u"tab_dock")
        self.dock_layout = QGridLayout(self.tab_dock)
        self.dock_layout.setSpacing(6)
        self.dock_layout.setContentsMargins(11, 11, 11, 11)
        self.dock_layout.setObjectName(u"dock_layout")
        self.dock_layout.setContentsMargins(2, 2, 2, 2)
        self.dock_type_combo = QComboBox(self.tab_dock)
        self.dock_type_combo.addItem("")
        self.dock_type_combo.addItem("")
        self.dock_type_combo.addItem("")
        self.dock_type_combo.addItem("")
        self.dock_type_combo.setObjectName(u"dock_type_combo")

        self.dock_layout.addWidget(self.dock_type_combo, 2, 1, 1, 1)

        self.dock_type_label = QLabel(self.tab_dock)
        self.dock_type_label.setObjectName(u"dock_type_label")

        self.dock_layout.addWidget(self.dock_type_label, 2, 0, 1, 1)

        self.dock_index_label = QLabel(self.tab_dock)
        self.dock_index_label.setObjectName(u"dock_index_label")

        self.dock_layout.addWidget(self.dock_index_label, 0, 0, 1, 1)

        self.dock_weakness_combo = QComboBox(self.tab_dock)
        self.dock_weakness_combo.setObjectName(u"dock_weakness_combo")

        self.dock_layout.addWidget(self.dock_weakness_combo, 3, 0, 1, 2)

        self.dock_index_spin = QSpinBox(self.tab_dock)
        self.dock_index_spin.setObjectName(u"dock_index_spin")

        self.dock_layout.addWidget(self.dock_index_spin, 0, 1, 1, 1)

        self.dock_connection_group = QGroupBox(self.tab_dock)
        self.dock_connection_group.setObjectName(u"dock_connection_group")
        self.dock_connection_layout = QHBoxLayout(self.dock_connection_group)
        self.dock_connection_layout.setSpacing(6)
        self.dock_connection_layout.setContentsMargins(11, 11, 11, 11)
        self.dock_connection_layout.setObjectName(u"dock_connection_layout")
        self.dock_connection_layout.setContentsMargins(2, 2, 2, 2)
        self.dock_connection_area_combo = QComboBox(self.dock_connection_group)
        self.dock_connection_area_combo.setObjectName(u"dock_connection_area_combo")

        self.dock_connection_layout.addWidget(self.dock_connection_area_combo)

        self.dock_connection_node_combo = QComboBox(self.dock_connection_group)
        self.dock_connection_node_combo.setObjectName(u"dock_connection_node_combo")

        self.dock_connection_layout.addWidget(self.dock_connection_node_combo)

        self.dock_connection_index_raw_spin = QSpinBox(self.dock_connection_group)
        self.dock_connection_index_raw_spin.setObjectName(u"dock_connection_index_raw_spin")
        self.dock_connection_index_raw_spin.setMaximumSize(QSize(40, 16777215))

        self.dock_connection_layout.addWidget(self.dock_connection_index_raw_spin)


        self.dock_layout.addWidget(self.dock_connection_group, 1, 0, 1, 2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.dock_layout.addItem(self.verticalSpacer, 4, 0, 1, 1)

        self.tab_widget.addTab(self.tab_dock, "")
        self.tab_pickup = QWidget()
        self.tab_pickup.setObjectName(u"tab_pickup")
        self.pickup_layout = QGridLayout(self.tab_pickup)
        self.pickup_layout.setSpacing(6)
        self.pickup_layout.setContentsMargins(11, 11, 11, 11)
        self.pickup_layout.setObjectName(u"pickup_layout")
        self.pickup_layout.setContentsMargins(2, 2, 2, 2)
        self.major_location_check = QCheckBox(self.tab_pickup)
        self.major_location_check.setObjectName(u"major_location_check")

        self.pickup_layout.addWidget(self.major_location_check, 1, 0, 1, 2)

        self.pickup_index_spin = QSpinBox(self.tab_pickup)
        self.pickup_index_spin.setObjectName(u"pickup_index_spin")
        self.pickup_index_spin.setMaximum(999)

        self.pickup_layout.addWidget(self.pickup_index_spin, 0, 1, 1, 1)

        self.pickup_index_label = QLabel(self.tab_pickup)
        self.pickup_index_label.setObjectName(u"pickup_index_label")

        self.pickup_layout.addWidget(self.pickup_index_label, 0, 0, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.pickup_layout.addItem(self.verticalSpacer_2, 2, 0, 1, 1)

        self.tab_widget.addTab(self.tab_pickup, "")
        self.tab_teleporter = QWidget()
        self.tab_teleporter.setObjectName(u"tab_teleporter")
        self.teleporter_layout = QGridLayout(self.tab_teleporter)
        self.teleporter_layout.setSpacing(6)
        self.teleporter_layout.setContentsMargins(11, 11, 11, 11)
        self.teleporter_layout.setObjectName(u"teleporter_layout")
        self.teleporter_layout.setContentsMargins(2, 2, 2, 2)
        self.teleporter_editable_check = QCheckBox(self.tab_teleporter)
        self.teleporter_editable_check.setObjectName(u"teleporter_editable_check")

        self.teleporter_layout.addWidget(self.teleporter_editable_check, 3, 0, 1, 1)

        self.teleporter_instance_id_edit = QLineEdit(self.tab_teleporter)
        self.teleporter_instance_id_edit.setObjectName(u"teleporter_instance_id_edit")

        self.teleporter_layout.addWidget(self.teleporter_instance_id_edit, 0, 1, 1, 1)

        self.teleporter_scan_asset_id_edit = QLineEdit(self.tab_teleporter)
        self.teleporter_scan_asset_id_edit.setObjectName(u"teleporter_scan_asset_id_edit")

        self.teleporter_layout.addWidget(self.teleporter_scan_asset_id_edit, 2, 1, 1, 1)

        self.teleporter_destination_group = QGroupBox(self.tab_teleporter)
        self.teleporter_destination_group.setObjectName(u"teleporter_destination_group")
        sizePolicy.setHeightForWidth(self.teleporter_destination_group.sizePolicy().hasHeightForWidth())
        self.teleporter_destination_group.setSizePolicy(sizePolicy)
        self.teleporter_destination_layout = QHBoxLayout(self.teleporter_destination_group)
        self.teleporter_destination_layout.setSpacing(6)
        self.teleporter_destination_layout.setContentsMargins(11, 11, 11, 11)
        self.teleporter_destination_layout.setObjectName(u"teleporter_destination_layout")
        self.teleporter_destination_layout.setContentsMargins(2, 2, 2, 2)
        self.teleporter_destination_world_combo = QComboBox(self.teleporter_destination_group)
        self.teleporter_destination_world_combo.setObjectName(u"teleporter_destination_world_combo")

        self.teleporter_destination_layout.addWidget(self.teleporter_destination_world_combo)

        self.teleporter_destination_area_combo = QComboBox(self.teleporter_destination_group)
        self.teleporter_destination_area_combo.setObjectName(u"teleporter_destination_area_combo")

        self.teleporter_destination_layout.addWidget(self.teleporter_destination_area_combo)


        self.teleporter_layout.addWidget(self.teleporter_destination_group, 1, 0, 1, 2)

        self.teleporter_scan_asset_id_label = QLabel(self.tab_teleporter)
        self.teleporter_scan_asset_id_label.setObjectName(u"teleporter_scan_asset_id_label")

        self.teleporter_layout.addWidget(self.teleporter_scan_asset_id_label, 2, 0, 1, 1)

        self.teleporter_vanilla_name_edit = QCheckBox(self.tab_teleporter)
        self.teleporter_vanilla_name_edit.setObjectName(u"teleporter_vanilla_name_edit")

        self.teleporter_layout.addWidget(self.teleporter_vanilla_name_edit, 3, 1, 1, 1)

        self.teleporter_instance_id_label = QLabel(self.tab_teleporter)
        self.teleporter_instance_id_label.setObjectName(u"teleporter_instance_id_label")

        self.teleporter_layout.addWidget(self.teleporter_instance_id_label, 0, 0, 1, 1)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.teleporter_layout.addItem(self.verticalSpacer_3, 4, 0, 1, 1)

        self.tab_widget.addTab(self.tab_teleporter, "")
        self.tab_event = QWidget()
        self.tab_event.setObjectName(u"tab_event")
        self.event_layout = QGridLayout(self.tab_event)
        self.event_layout.setSpacing(6)
        self.event_layout.setContentsMargins(11, 11, 11, 11)
        self.event_layout.setObjectName(u"event_layout")
        self.event_layout.setContentsMargins(2, 2, 2, 2)
        self.event_resource_label = QLabel(self.tab_event)
        self.event_resource_label.setObjectName(u"event_resource_label")

        self.event_layout.addWidget(self.event_resource_label, 0, 0, 1, 1)

        self.event_resource_combo = QComboBox(self.tab_event)
        self.event_resource_combo.setObjectName(u"event_resource_combo")

        self.event_layout.addWidget(self.event_resource_combo, 0, 1, 1, 1)

        self.verticalSpacer_4 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.event_layout.addItem(self.verticalSpacer_4, 1, 0, 1, 1)

        self.tab_widget.addTab(self.tab_event, "")
        self.tab_translator_gate = QWidget()
        self.tab_translator_gate.setObjectName(u"tab_translator_gate")
        self.translator_gate_layout = QGridLayout(self.tab_translator_gate)
        self.translator_gate_layout.setSpacing(6)
        self.translator_gate_layout.setContentsMargins(11, 11, 11, 11)
        self.translator_gate_layout.setObjectName(u"translator_gate_layout")
        self.translator_gate_layout.setContentsMargins(2, 2, 2, 2)
        self.translator_gate_spin = QSpinBox(self.tab_translator_gate)
        self.translator_gate_spin.setObjectName(u"translator_gate_spin")
        self.translator_gate_spin.setMaximum(999)

        self.translator_gate_layout.addWidget(self.translator_gate_spin, 0, 1, 1, 1)

        self.translator_gate_label = QLabel(self.tab_translator_gate)
        self.translator_gate_label.setObjectName(u"translator_gate_label")

        self.translator_gate_layout.addWidget(self.translator_gate_label, 0, 0, 1, 1)

        self.verticalSpacer_5 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.translator_gate_layout.addItem(self.verticalSpacer_5, 1, 0, 1, 1)

        self.tab_widget.addTab(self.tab_translator_gate, "")
        self.tab_logbook = QWidget()
        self.tab_logbook.setObjectName(u"tab_logbook")
        self.logbook_layout = QGridLayout(self.tab_logbook)
        self.logbook_layout.setSpacing(6)
        self.logbook_layout.setContentsMargins(11, 11, 11, 11)
        self.logbook_layout.setObjectName(u"logbook_layout")
        self.logbook_layout.setContentsMargins(2, 2, 2, 2)
        self.logbook_extra_label = QLabel(self.tab_logbook)
        self.logbook_extra_label.setObjectName(u"logbook_extra_label")

        self.logbook_layout.addWidget(self.logbook_extra_label, 2, 0, 1, 1)

        self.lore_type_combo = QComboBox(self.tab_logbook)
        self.lore_type_combo.addItem("")
        self.lore_type_combo.addItem("")
        self.lore_type_combo.addItem("")
        self.lore_type_combo.addItem("")
        self.lore_type_combo.setObjectName(u"lore_type_combo")

        self.logbook_layout.addWidget(self.lore_type_combo, 1, 1, 1, 1)

        self.logbook_extra_combo = QComboBox(self.tab_logbook)
        self.logbook_extra_combo.setObjectName(u"logbook_extra_combo")

        self.logbook_layout.addWidget(self.logbook_extra_combo, 2, 1, 1, 1)

        self.logbook_string_asset_id_edit = QLineEdit(self.tab_logbook)
        self.logbook_string_asset_id_edit.setObjectName(u"logbook_string_asset_id_edit")

        self.logbook_layout.addWidget(self.logbook_string_asset_id_edit, 0, 1, 1, 1)

        self.logbook_string_asset_id_label = QLabel(self.tab_logbook)
        self.logbook_string_asset_id_label.setObjectName(u"logbook_string_asset_id_label")

        self.logbook_layout.addWidget(self.logbook_string_asset_id_label, 0, 0, 1, 1)

        self.lore_type_label = QLabel(self.tab_logbook)
        self.lore_type_label.setObjectName(u"lore_type_label")

        self.logbook_layout.addWidget(self.lore_type_label, 1, 0, 1, 1)

        self.verticalSpacer_6 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.logbook_layout.addItem(self.verticalSpacer_6, 3, 0, 1, 1)

        self.tab_widget.addTab(self.tab_logbook, "")
        self.tab_player_ship = QWidget()
        self.tab_player_ship.setObjectName(u"tab_player_ship")
        self.player_ship_layout = QVBoxLayout(self.tab_player_ship)
        self.player_ship_layout.setSpacing(6)
        self.player_ship_layout.setContentsMargins(11, 11, 11, 11)
        self.player_ship_layout.setObjectName(u"player_ship_layout")
        self.player_ship_layout.setContentsMargins(2, 2, 2, 2)
        self.player_ship_unlocked_button = QPushButton(self.tab_player_ship)
        self.player_ship_unlocked_button.setObjectName(u"player_ship_unlocked_button")

        self.player_ship_layout.addWidget(self.player_ship_unlocked_button)

        self.player_ship_unlocked_group = QGroupBox(self.tab_player_ship)
        self.player_ship_unlocked_group.setObjectName(u"player_ship_unlocked_group")
        self.player_ship_unlocked_layout = QGridLayout(self.player_ship_unlocked_group)
        self.player_ship_unlocked_layout.setSpacing(2)
        self.player_ship_unlocked_layout.setContentsMargins(11, 11, 11, 11)
        self.player_ship_unlocked_layout.setObjectName(u"player_ship_unlocked_layout")
        self.player_ship_unlocked_layout.setContentsMargins(2, 2, 2, 2)

        self.player_ship_layout.addWidget(self.player_ship_unlocked_group)

        self.tab_widget.addTab(self.tab_player_ship, "")

        self.main_layout.addWidget(self.tab_widget)

        self.verticalSpacer_7 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.main_layout.addItem(self.verticalSpacer_7)

        self.button_box = QDialogButtonBox(NodeDetailsPopup)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Save)

        self.main_layout.addWidget(self.button_box)


        self.retranslateUi(NodeDetailsPopup)

        self.tab_widget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(NodeDetailsPopup)
    # setupUi

    def retranslateUi(self, NodeDetailsPopup):
        NodeDetailsPopup.setWindowTitle(QCoreApplication.translate("NodeDetailsPopup", u"Item Configuration", None))
        self.name_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Name:", None))
        self.heals_check.setText(QCoreApplication.translate("NodeDetailsPopup", u"Heals", None))
        self.location_group.setTitle(QCoreApplication.translate("NodeDetailsPopup", u"Location", None))
        self.location_x_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"X", None))
        self.location_y_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Y", None))
        self.location_z_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Z", None))
        self.node_type_combo.setItemText(0, QCoreApplication.translate("NodeDetailsPopup", u"Generic", None))
        self.node_type_combo.setItemText(1, QCoreApplication.translate("NodeDetailsPopup", u"Dock", None))
        self.node_type_combo.setItemText(2, QCoreApplication.translate("NodeDetailsPopup", u"Pickup", None))
        self.node_type_combo.setItemText(3, QCoreApplication.translate("NodeDetailsPopup", u"Teleporter", None))
        self.node_type_combo.setItemText(4, QCoreApplication.translate("NodeDetailsPopup", u"Event", None))
        self.node_type_combo.setItemText(5, QCoreApplication.translate("NodeDetailsPopup", u"Translator Gate", None))
        self.node_type_combo.setItemText(6, QCoreApplication.translate("NodeDetailsPopup", u"Logbook", None))
        self.node_type_combo.setItemText(7, QCoreApplication.translate("NodeDetailsPopup", u"Player Ship", None))

        self.label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Nothing!", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_generic), QCoreApplication.translate("NodeDetailsPopup", u"Generic", None))
        self.dock_type_combo.setItemText(0, QCoreApplication.translate("NodeDetailsPopup", u"Door", None))
        self.dock_type_combo.setItemText(1, QCoreApplication.translate("NodeDetailsPopup", u"Morph Ball Door", None))
        self.dock_type_combo.setItemText(2, QCoreApplication.translate("NodeDetailsPopup", u"Other", None))
        self.dock_type_combo.setItemText(3, QCoreApplication.translate("NodeDetailsPopup", u"Portal", None))

        self.dock_type_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Dock type:", None))
        self.dock_index_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Dock index:", None))
        self.dock_connection_group.setTitle(QCoreApplication.translate("NodeDetailsPopup", u"Dock Connection", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_dock), QCoreApplication.translate("NodeDetailsPopup", u"Dock", None))
        self.major_location_check.setText(QCoreApplication.translate("NodeDetailsPopup", u"Is major location", None))
        self.pickup_index_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Pickup Index:", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_pickup), QCoreApplication.translate("NodeDetailsPopup", u"Pickup", None))
        self.teleporter_editable_check.setText(QCoreApplication.translate("NodeDetailsPopup", u"Randomizable?", None))
        self.teleporter_destination_group.setTitle(QCoreApplication.translate("NodeDetailsPopup", u"World/Area Selection", None))
        self.teleporter_scan_asset_id_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Scan asset id:", None))
        self.teleporter_vanilla_name_edit.setText(QCoreApplication.translate("NodeDetailsPopup", u"Keep name when vanilla?", None))
        self.teleporter_instance_id_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Instance id:", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_teleporter), QCoreApplication.translate("NodeDetailsPopup", u"Teleporter", None))
        self.event_resource_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Event:", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_event), QCoreApplication.translate("NodeDetailsPopup", u"Event", None))
        self.translator_gate_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Gate Index", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_translator_gate), QCoreApplication.translate("NodeDetailsPopup", u"Translator Gate", None))
        self.logbook_extra_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Extra:", None))
        self.lore_type_combo.setItemText(0, QCoreApplication.translate("NodeDetailsPopup", u"Luminoth Lore", None))
        self.lore_type_combo.setItemText(1, QCoreApplication.translate("NodeDetailsPopup", u"Luminoth Warrior", None))
        self.lore_type_combo.setItemText(2, QCoreApplication.translate("NodeDetailsPopup", u"Pirate Lore", None))
        self.lore_type_combo.setItemText(3, QCoreApplication.translate("NodeDetailsPopup", u"Sky Temple Key Hint", None))

        self.logbook_string_asset_id_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"String asset id", None))
        self.lore_type_label.setText(QCoreApplication.translate("NodeDetailsPopup", u"Lore type:", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_logbook), QCoreApplication.translate("NodeDetailsPopup", u"Logbook", None))
        self.player_ship_unlocked_button.setText(QCoreApplication.translate("NodeDetailsPopup", u"Edit unlocked by", None))
        self.player_ship_unlocked_group.setTitle("")
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_player_ship), QCoreApplication.translate("NodeDetailsPopup", u"Player Ship", None))
    # retranslateUi




from typing import List, Dict

from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources, add_resource_gain_to_current_resources
from randovania.generator.item_pool import PoolResults
from randovania.generator.item_pool.pickup_creator import create_sky_temple_key
from randovania.layout.echoes_configuration import LayoutSkyTempleKeyMode
from randovania.resolver.exceptions import InvalidConfiguration


def add_sky_temple_key_distribution_logic(resource_database: ResourceDatabase,
                                          mode: LayoutSkyTempleKeyMode,
                                          ) -> PoolResults:
    """
    Adds the given Sky Temple Keys to the item pool
    :param resource_database:
    :param mode:
    :return:
    """

    item_pool: List[PickupEntry] = []
    new_assignment: Dict[PickupIndex, PickupEntry] = {}
    initial_resources: CurrentResources = {}

    if mode == LayoutSkyTempleKeyMode.ALL_BOSSES or mode == LayoutSkyTempleKeyMode.ALL_GUARDIANS:
        locations_to_place = _GUARDIAN_INDICES[:]
        if mode == LayoutSkyTempleKeyMode.ALL_BOSSES:
            locations_to_place += _SUB_GUARDIAN_INDICES

        for key_number, location in enumerate(locations_to_place):
            new_assignment[location] = create_sky_temple_key(key_number, resource_database)
        first_automatic_key = len(locations_to_place)

    else:
        keys_to_place = mode.value
        if not isinstance(keys_to_place, int):
            raise InvalidConfiguration("Unknown Sky Temple Key mode: {}".format(mode))

        for key_number in range(keys_to_place):
            item_pool.append(create_sky_temple_key(key_number, resource_database))
        first_automatic_key = keys_to_place

    for automatic_key_number in range(first_automatic_key, 9):
        add_resource_gain_to_current_resources(
            create_sky_temple_key(automatic_key_number, resource_database).all_resources,
            initial_resources
        )

    return PoolResults(item_pool, new_assignment, initial_resources)


_GUARDIAN_INDICES = [
    PickupIndex(43),  # Dark Suit
    PickupIndex(79),  # Dark Visor
    PickupIndex(115),  # Annihilator Beam
]
_SUB_GUARDIAN_INDICES = [
    PickupIndex(38),  # Morph Ball Bomb
    PickupIndex(37),  # Space Jump Boots
    PickupIndex(75),  # Boost Ball
    PickupIndex(86),  # Grapple Beam
    PickupIndex(102),  # Spider Ball
    PickupIndex(88),  # Main Power Bombs
]



import contextlib
import datetime

import pytest
from PySide2 import QtWidgets
from PySide2.QtWidgets import QMessageBox
from mock import MagicMock, AsyncMock, ANY

from randovania.generator import base_patches_factory
from randovania.gui.game_session_window import GameSessionWindow
from randovania.layout.permalink import Permalink
from randovania.network_client.game_session import GameSessionEntry, PlayerSessionEntry, User, GameSessionAction
from randovania.network_common.admin_actions import SessionAdminGlobalAction
from randovania.network_common.session_state import GameSessionState


@pytest.fixture(name="window")
def _window(skip_qtbot):
    game_connection = MagicMock()
    game_connection.backend.lock_identifier = None
    game_connection.pretty_current_status = "Pretty Status"
    window = GameSessionWindow(MagicMock(), game_connection, MagicMock(), MagicMock(), MagicMock())
    return window


@pytest.mark.asyncio
async def test_on_game_session_updated(preset_manager, skip_qtbot):
    # Setup
    network_client = MagicMock()
    network_client.current_user = User(id=12, name="Player A")
    network_client.session_self_update = AsyncMock()
    game_connection = MagicMock()
    game_connection.pretty_current_status = "Maybe Connected"
    game_connection.backend.lock_identifier = None

    initial_session = GameSessionEntry(
        id=1234,
        name="The Session",
        presets=[preset_manager.default_preset, preset_manager.default_preset],
        players={
            12: PlayerSessionEntry(12, "Player A", 0, True, "Online"),
        },
        actions=[],
        seed_hash=None,
        word_hash=None,
        spoiler=None,
        permalink=None,
        state=GameSessionState.SETUP,
        generation_in_progress=None,
    )
    second_session = GameSessionEntry(
        id=1234,
        name="The Session",
        presets=[preset_manager.default_preset],
        players={
            12: PlayerSessionEntry(12, "Player A", 0, True, "Online"),
            24: PlayerSessionEntry(24, "Player B", None, False, "Online"),
        },
        actions=[
            GameSessionAction("Hello", datetime.datetime(year=2020, month=1, day=5))
        ],
        seed_hash="AB12",
        word_hash="Chykka Required",
        spoiler=True,
        permalink="<permalink>",
        state=GameSessionState.IN_PROGRESS,
        generation_in_progress=None,
    )
    network_client.current_game_session = initial_session

    window = await GameSessionWindow.create_and_update(network_client, game_connection, preset_manager,
                                                       MagicMock(), MagicMock())
    window.update_multiworld_client_status = AsyncMock()

    # Run
    await window.on_game_session_updated(second_session)
    window.update_multiworld_client_status.assert_awaited()
    network_client.session_self_update.assert_awaited_once_with(
        game_connection.get_current_inventory.return_value,
        game_connection.current_status,
        game_connection.backend.backend_choice,
    )


@pytest.mark.parametrize("in_game", [False, True])
@pytest.mark.asyncio
async def test_update_multiworld_client_status(window, in_game):
    # Setup
    window._game_session = MagicMock()
    window._game_session.state = GameSessionState.IN_PROGRESS if in_game else GameSessionState.SETUP
    window.multiworld_client = MagicMock()
    window.multiworld_client.is_active = not in_game
    window.multiworld_client.start = AsyncMock()
    window.multiworld_client.stop = AsyncMock()

    # Run
    await window.update_multiworld_client_status()

    # Assert
    if in_game:
        called, not_called = window.multiworld_client.start, window.multiworld_client.stop
    else:
        called, not_called = window.multiworld_client.stop, window.multiworld_client.start
    called.assert_awaited_once()
    not_called.assert_not_awaited()


@pytest.mark.asyncio
async def test_row_show_preset_summary(window, mocker, preset_manager):
    # Setup
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)

    row = MagicMock()
    window.rows = [row]
    window._game_session = MagicMock()
    window._game_session.presets = [preset_manager.default_preset]

    # Run
    await window._row_show_preset_summary(row)

    # Assert
    execute_dialog.assert_awaited_once()


@pytest.mark.parametrize(["has_background_process", "generation_in_progress", "seed_hash", "expected_text"], [
    (True, None, None, "Stop"),
    (False, True, None, "Abort generation"),
    (False, None, None, "Generate game"),
    (False, None, True, "Clear generated game"),
])
def test_update_background_process_button(window, has_background_process, generation_in_progress, seed_hash,
                                          expected_text):
    window._game_session = MagicMock()

    window._background_thread = True if has_background_process else None
    window._game_session.generation_in_progress = generation_in_progress
    window._game_session.seed_hash = seed_hash

    # Run
    window.update_background_process_button()

    # Assert
    assert window.background_process_button.text() == expected_text


def test_sync_background_process_to_game_session_other_generation(window):
    window._game_session = MagicMock()
    window._game_session.generation_in_progress = True
    window._generating_game = False

    # Run
    window.sync_background_process_to_game_session()

    # Assert
    assert window.progress_label.text().startswith("Game being generated by")


def test_sync_background_process_to_game_session_stop_background(window):
    window._game_session = MagicMock()
    window._game_session.generation_in_progress = None
    window._background_thread = True
    window._generating_game = True
    window.stop_background_process = MagicMock()

    # Run
    window.sync_background_process_to_game_session()

    # Assert
    window.stop_background_process.assert_called_once_with()


def test_sync_background_process_to_game_session_nothing(window):
    window._game_session = MagicMock()
    window._game_session.generation_in_progress = None
    window._background_thread = None
    window._generating_game = False
    window.stop_background_process = MagicMock()

    # Run
    window.sync_background_process_to_game_session()

    # Assert
    assert window.progress_label.text() == ""


@pytest.mark.parametrize("has_game", [False, True])
@pytest.mark.asyncio
async def test_update_logic_settings_window(window, mocker, has_game):
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)

    window._game_session = MagicMock()
    window._logic_settings_window = MagicMock()
    window._game_session.seed_hash = True if has_game else None

    # Run
    await window.update_logic_settings_window()

    # Assert
    if has_game:
        window._logic_settings_window.setEnabled.assert_not_called()
        window._logic_settings_window.reject.assert_called_once_with()
        execute_dialog.assert_awaited_once()
    else:
        window._logic_settings_window.setEnabled.assert_called_once()
        execute_dialog.assert_not_awaited()


@pytest.mark.asyncio
async def test_change_password(window, mocker):
    def set_text_value(dialog: QtWidgets.QInputDialog):
        dialog.setTextValue("magoo")
        return QtWidgets.QDialog.Accepted

    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                  side_effect=set_text_value)
    window._admin_global_action = AsyncMock()

    # Run
    await window.change_password()

    # Assert
    execute_dialog.assert_awaited_once()
    window._admin_global_action.assert_awaited_once_with(SessionAdminGlobalAction.CHANGE_PASSWORD, "magoo")


@pytest.mark.asyncio
async def test_generate_game(window, mocker, preset_manager):
    mock_generate_layout: MagicMock = mocker.patch("randovania.interface_common.simplified_patcher.generate_layout")
    mock_randint: MagicMock = mocker.patch("random.randint", return_value=5000)

    spoiler = True
    game_session = MagicMock()
    game_session.presets = [preset_manager.default_preset, preset_manager.default_preset]

    window._game_session = game_session
    window._upload_layout_description = AsyncMock()
    window._admin_global_action = AsyncMock()

    # Run
    await window.generate_game(spoiler)

    # Assert
    mock_randint.assert_called_once_with(0, 2 ** 31)
    mock_generate_layout.assert_called_once_with(
        progress_update=ANY,
        permalink=Permalink(
            seed_number=mock_randint.return_value,
            spoiler=spoiler,
            presets={
                0: preset_manager.default_preset.get_preset(),
                1: preset_manager.default_preset.get_preset(),
            },
        ),
        options=window._options
    )
    window._upload_layout_description.assert_awaited_once_with(mock_generate_layout.return_value)


@pytest.mark.asyncio
async def test_check_dangerous_presets(window, mocker):
    mock_warning = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)
    mock_warning.return_value = QtWidgets.QMessageBox.No

    game_session = MagicMock()
    game_session.presets = [MagicMock(), MagicMock(), MagicMock()]
    game_session.presets[0].name = "Preset A"
    game_session.presets[0].dangerous_settings.return_value = ["Cake"]
    game_session.presets[1].name = "Preset B"
    game_session.presets[1].dangerous_settings.return_value = ["Bomb", "Knife"]
    game_session.presets[2].dangerous_settings.return_value = []

    window._game_session = game_session
    window.team_players = [MagicMock(), MagicMock()]
    window.team_players[0].player.name = "Crazy Person"
    window.team_players[1].player = None

    permalink = MagicMock()
    permalink.presets = {i: preset for i, preset in enumerate(game_session.presets)}

    # Run
    result = await window._check_dangerous_presets(permalink)

    # Assert
    message = ("The following presets have settings that can cause an impossible game:\n"
               "\nCrazy Person - Preset A: Cake"
               "\nPlayer 2 - Preset B: Bomb, Knife"
               "\n\nDo you want to continue?")
    mock_warning.assert_awaited_once_with(window, "Dangerous preset", message,
                                          QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
    assert not result


@pytest.mark.asyncio
async def test_copy_permalink(window, mocker):
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
    game_session = MagicMock()
    game_session.permalink = "<permalink>"

    window._game_session = game_session

    # Run
    await window.copy_permalink()

    # Assert
    execute_dialog.assert_awaited_once()
    assert execute_dialog.call_args.args[0].textValue() == "<permalink>"


@pytest.mark.asyncio
async def test_import_permalink(window, mocker):
    mock_permalink_dialog = mocker.patch("randovania.gui.game_session_window.PermalinkDialog")
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
    execute_dialog.return_value = QtWidgets.QDialog.Accepted
    mock_warning = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)
    mock_warning.return_value = QtWidgets.QMessageBox.Yes

    permalink = mock_permalink_dialog.return_value.get_permalink_from_field.return_value
    permalink.player_count = 2
    permalink.presets = {0: MagicMock(), 1: MagicMock()}
    permalink.presets[0].is_same_configuration.return_value = False

    game_session = MagicMock()
    game_session.num_rows = 2
    game_session.presets = [MagicMock(), MagicMock()]

    window._game_session = game_session
    window.generate_game_with_permalink = AsyncMock()

    # Run
    await window.import_permalink()

    # Assert
    execute_dialog.assert_awaited_once_with(mock_permalink_dialog.return_value)
    mock_warning.assert_awaited_once_with(window, "Different presets", ANY, ANY, QtWidgets.QMessageBox.No)
    window.generate_game_with_permalink.assert_awaited_once_with(permalink)


@pytest.mark.parametrize(["expecting_kick", "already_kicked"], [
    (False, True),
    (False, False),
    (True, False),
])
@pytest.mark.asyncio
async def test_on_kicked(skip_qtbot, window, mocker, expecting_kick, already_kicked):
    mock_warning = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)

    window.network_client.leave_game_session = AsyncMock()
    window._game_session = MagicMock()
    window._already_kicked = already_kicked
    window._expecting_kick = expecting_kick
    window.close = MagicMock(return_value=None)

    # Run
    await window._on_kicked()
    if not already_kicked:
        skip_qtbot.waitUntil(window.close)

    # Assert
    if already_kicked:
        window.network_client.leave_game_session.assert_not_awaited()
        mock_warning.assert_not_awaited()
        window.close.assert_not_called()
    else:
        window.network_client.leave_game_session.assert_awaited_once_with(False)
        if expecting_kick:
            mock_warning.assert_not_awaited()
        else:
            mock_warning.assert_awaited_once()
        window.close.assert_called_once_with()


@pytest.mark.parametrize("exception", [False, True])
@pytest.mark.parametrize("accept", [False, True])
@pytest.mark.asyncio
async def test_delete_session(window, mocker, accept, exception):
    mock_warning = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)
    mock_warning.return_value = QtWidgets.QMessageBox.Yes if accept else QtWidgets.QMessageBox.No

    window.network_client.session_admin_global = AsyncMock()
    if exception and accept:
        window.network_client.session_admin_global.side_effect = RuntimeError("error")
        expectation = pytest.raises(RuntimeError, match="error")
    else:
        expectation = contextlib.nullcontext()

    # Run
    with expectation:
        await window.delete_session()

    # Assert
    mock_warning.assert_awaited_once()
    if accept:
        window.network_client.session_admin_global.assert_awaited_once_with(SessionAdminGlobalAction.DELETE_SESSION,
                                                                            None)
        assert window._expecting_kick != exception
    else:
        window.network_client.session_admin_global.assert_not_awaited()
        assert not window._expecting_kick


@pytest.mark.parametrize("accept", [False, True])
@pytest.mark.asyncio
async def test_finish_session(window, accept, mocker):
    mock_warning = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)
    mock_warning.return_value = QtWidgets.QMessageBox.Yes if accept else QtWidgets.QMessageBox.No

    window.network_client.session_admin_global = AsyncMock()

    # Run
    await window.finish_session()

    # Assert
    mock_warning.assert_awaited_once()
    if accept:
        window.network_client.session_admin_global.assert_awaited_once_with(SessionAdminGlobalAction.FINISH_SESSION,
                                                                            None)
    else:
        window.network_client.session_admin_global.assert_not_awaited()


@pytest.mark.asyncio
async def test_save_iso(window, mocker, preset_manager, echoes_game_description):
    mock_input_dialog = mocker.patch("randovania.gui.game_session_window.GameInputDialog")
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=QtWidgets.QDialog.Accepted)
    mock_unpack_iso: MagicMock = mocker.patch("randovania.interface_common.simplified_patcher.unpack_iso")
    mock_apply_patcher: MagicMock = mocker.patch("randovania.interface_common.simplified_patcher.apply_patcher_file")
    mock_pack_iso: MagicMock = mocker.patch("randovania.interface_common.simplified_patcher.pack_iso")

    window._game_session = MagicMock()
    window._game_session.players[window.network_client.current_user.id].is_observer = False
    window._game_session.players[window.network_client.current_user.id].row = 0
    window._game_session.presets = {0: preset_manager.default_preset}
    layout_configuration = preset_manager.default_preset.get_preset().configuration
    window.network_client.session_admin_player = AsyncMock()

    def run_in_background_thread(work, *args):
        work(MagicMock())

    window.run_in_background_thread = MagicMock(side_effect=run_in_background_thread)

    # Run
    await window.save_iso()

    # Assert
    mock_execute_dialog.assert_awaited_once_with(mock_input_dialog.return_value)
    mock_unpack_iso.assert_called_once_with(
        input_iso=mock_input_dialog.return_value.input_file,
        options=window._options,
        progress_update=ANY
    )
    mock_apply_patcher.assert_called_once_with(
        patcher_file=window.network_client.session_admin_player.return_value,
        game_specific=base_patches_factory.create_game_specific(layout_configuration, echoes_game_description),
        shareable_hash=window._game_session.seed_hash,
        options=window._options,
        progress_update=ANY,
    )
    mock_pack_iso.assert_called_once_with(output_iso=mock_input_dialog.return_value.output_file,
                                          options=window._options,
                                          progress_update=ANY)


@pytest.mark.parametrize("dialog_response", [QMessageBox.Yes, QMessageBox.No, QMessageBox.Cancel])
@pytest.mark.parametrize("is_member", [False, True])
@pytest.mark.asyncio
async def test_on_close_event(window: GameSessionWindow, mocker, dialog_response, is_member):
    # Setup
    execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock,
                                  return_value=dialog_response)
    super_close_event = mocker.patch("PySide2.QtWidgets.QMainWindow.closeEvent")
    event = MagicMock()
    window._game_session = MagicMock()
    window._game_session.players = [window.network_client.current_user.id] if is_member else []
    window.network_client.leave_game_session = AsyncMock()
    window.network_client.connection_state.is_disconnected = dialog_response == QMessageBox.Yes

    # Run
    await window._on_close_event(event)

    if is_member:
        execute_dialog.assert_awaited_once()
    else:
        execute_dialog.assert_not_awaited()

    if is_member and dialog_response == QMessageBox.Cancel:
        event.ignore.assert_called_once_with()
        super_close_event.assert_not_called()
    else:
        event.ignore.assert_not_called()
        super_close_event.assert_called_once_with(event)

    if is_member and dialog_response != QMessageBox.Cancel:
        window.network_client.leave_game_session.assert_awaited_once_with(dialog_response == QMessageBox.Yes)
    else:
        window.network_client.leave_game_session.assert_not_awaited()



import math
import multiprocessing
import time
from argparse import ArgumentParser
from pathlib import Path

from randovania.cli import echoes_lib
from randovania.generator import generator
from randovania.interface_common import sleep_inhibitor
from randovania.layout.permalink import Permalink


def batch_distribute_helper(base_permalink: Permalink,
                            seed_number: int,
                            timeout: int,
                            validate: bool,
                            output_dir: Path,
                            ) -> float:
    permalink = Permalink(
        seed_number=seed_number,
        spoiler=True,
        presets=base_permalink.presets,
    )

    start_time = time.perf_counter()
    description = generator.generate_description(permalink=permalink, status_update=None,
                                                 validate_after_generation=validate, timeout=timeout,
                                                 attempts=0)
    delta_time = time.perf_counter() - start_time

    description.save_to_file(output_dir.joinpath("{}.json".format(seed_number)))
    return delta_time


def batch_distribute_command_logic(args):
    finished_count = 0

    timeout: int = args.timeout
    validate: bool = args.validate

    output_dir: Path = args.output_dir
    output_dir.mkdir(parents=True, exist_ok=True)

    seed_count = args.seed_count
    num_digits = math.ceil(math.log10(seed_count + 1))
    number_format = "[{0:" + str(num_digits) + "d}/{1}] "
    base_permalink = Permalink.from_str(args.permalink)

    def report_update(msg: str):
        nonlocal finished_count
        finished_count += 1
        print(number_format.format(finished_count, seed_count) + msg)

    def callback(result):
        report_update(f"Finished seed in {result} seconds.")

    def error_callback(e):
        report_update(f"Failed to generate seed: {e}")

    with multiprocessing.Pool(processes=args.process_count) as pool, sleep_inhibitor.get_inhibitor():
        for seed_number in range(base_permalink.seed_number, base_permalink.seed_number + args.seed_count):
            pool.apply_async(
                func=batch_distribute_helper,
                args=(base_permalink, seed_number, timeout, validate, output_dir),
                callback=callback,
                error_callback=error_callback,
            )
        pool.close()
        pool.join()


def add_batch_distribute_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "batch-distribute",
        help="Generate multiple seeds in parallel"
    )

    parser.add_argument("permalink", type=str, help="The permalink to use")
    parser.add_argument("--process-count", type=int, help="How many processes to use. Defaults to CPU count.")
    parser.add_argument(
        "--timeout",
        type=int,
        default=90,
        help="How many seconds to wait before timing out a generation/validation.")
    echoes_lib.add_validate_argument(parser)
    parser.add_argument(
        "seed_count",
        type=int,
        help="How many seeds to generate.")
    parser.add_argument(
        "output_dir",
        type=Path,
        help="Where to place the seed logs.")
    parser.set_defaults(func=batch_distribute_command_logic)



from typing import Dict, NamedTuple

from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.translator_gate import TranslatorGate


class PickupTarget(NamedTuple):
    pickup: PickupEntry
    player: int


PickupAssignment = Dict[PickupIndex, PickupTarget]
GateAssignment = Dict[TranslatorGate, ItemResourceInfo]



import dataclasses
from typing import Optional, Tuple

from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo


@dataclasses.dataclass(frozen=True)
class EchoesBeamConfiguration:
    item: ItemResourceInfo
    ammo_a: Optional[ItemResourceInfo]
    ammo_b: Optional[ItemResourceInfo]
    uncharged_cost: int
    charged_cost: int
    combo_missile_cost: int
    combo_ammo_cost: int


@dataclasses.dataclass(frozen=True)
class EchoesGameSpecific:
    energy_per_tank: float
    safe_zone_heal_per_second: float
    beam_configurations: Tuple[EchoesBeamConfiguration, ...]
    dangerous_energy_tank: bool



from typing import Union, Iterator, Tuple, TextIO

from randovania.game_description.area import Area
from randovania.game_description.dock import DockType
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import Node, DockNode, TeleporterNode, PickupNode, EventNode, TranslatorGateNode, \
    LogbookNode, LoreType, PlayerShipNode
from randovania.game_description.requirements import ResourceRequirement, RequirementAnd, RequirementOr, \
    RequirementTemplate, Requirement
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world_list import WorldList
from randovania.interface_common.enum_lib import iterate_enum
from randovania.layout.trick_level import LayoutTrickLevel


def pretty_print_resource_requirement(requirement: ResourceRequirement) -> str:
    if requirement.resource.resource_type == ResourceType.TRICK:
        return f"{requirement.resource} ({LayoutTrickLevel.from_number(requirement.amount).long_name})"
    else:
        return requirement.pretty_text


def pretty_print_requirement_array(requirement: Union[RequirementAnd, RequirementOr],
                                   level: int) -> Iterator[Tuple[int, str]]:
    if len(requirement.items) == 1:
        yield from pretty_print_requirement(requirement.items[0], level)
        return

    resource_requirements = [item for item in requirement.items if isinstance(item, ResourceRequirement)]
    template_requirements = [item for item in requirement.items if isinstance(item, RequirementTemplate)]
    other_requirements = [item for item in requirement.items if isinstance(item, (RequirementAnd, RequirementOr))]
    assert len(resource_requirements) + len(template_requirements) + len(other_requirements) == len(requirement.items)

    pretty_resources = [
        pretty_print_resource_requirement(item)
        for item in sorted(resource_requirements)
    ]
    sorted_templates = list(sorted(item.template_name for item in template_requirements))

    if isinstance(requirement, RequirementOr):
        title = "Any"
        combinator = " or "
    else:
        title = "All"
        combinator = " and "

    if len(other_requirements) == 0:
        yield level, combinator.join(pretty_resources + sorted_templates)
    else:
        yield level, f"{title} of the following:"
        if pretty_resources or sorted_templates:
            yield level + 1, combinator.join(pretty_resources + sorted_templates)
        for item in other_requirements:
            yield from pretty_print_requirement(item, level + 1)


def pretty_print_requirement(requirement: Requirement, level: int = 0) -> Iterator[Tuple[int, str]]:
    if requirement == Requirement.impossible():
        yield level, "Impossible"

    elif requirement == Requirement.trivial():
        yield level, "Trivial"

    elif isinstance(requirement, (RequirementAnd, RequirementOr)):
        yield from pretty_print_requirement_array(requirement, level)

    elif isinstance(requirement, ResourceRequirement):
        yield level, pretty_print_resource_requirement(requirement)

    elif isinstance(requirement, RequirementTemplate):
        yield level, requirement.template_name
    else:
        raise RuntimeError(f"Unknown requirement type: {type(requirement)} - {requirement}")


def pretty_print_node_type(node: Node, world_list: WorldList):
    if isinstance(node, DockNode):
        try:
            other = world_list.resolve_dock_connection(world_list.nodes_to_world(node), node.default_connection)
            other_name = world_list.node_name(other)
        except IndexError as e:
            other_name = (f"(Asset {node.default_connection.area_asset_id:x}, "
                          f"index {node.default_connection.dock_index}) [{e}]")

        return f"{node.default_dock_weakness.name} to {other_name}"

    elif isinstance(node, TeleporterNode):
        other = world_list.area_by_area_location(node.default_connection)
        return f"Teleporter to {world_list.area_name(other)}"

    elif isinstance(node, PickupNode):
        return f"Pickup {node.pickup_index.index}; Major Location? {node.major_location}"

    elif isinstance(node, EventNode):
        return f"Event {node.event.long_name}"

    elif isinstance(node, TranslatorGateNode):
        return f"Translator Gate ({node.gate})"

    elif isinstance(node, LogbookNode):
        message = ""
        if node.lore_type == LoreType.LUMINOTH_LORE:
            message = f" ({node.required_translator.long_name})"
        return f"Logbook {node.lore_type.long_name}{message} for {node.string_asset_id:x}"

    elif isinstance(node, PlayerShipNode):
        unlocked_pretty = list(pretty_print_requirement(node.is_unlocked))
        if len(unlocked_pretty) > 1:
            unlocked_by = "Complex requirement"
        else:
            unlocked_by = unlocked_pretty[0][1]
        return f"Player Ship (Unlocked by {unlocked_by})"

    return ""


def pretty_print_area(game: GameDescription, area: Area, print_function=print):
    print_function(area.name)
    print_function("Asset id: {}".format(area.area_asset_id))
    for i, node in enumerate(area.nodes):
        message = f"> {node.name}; Heals? {node.heal}"
        if area.default_node_index == i:
            message += "; Spawn Point"
        print_function(message)

        description_line = pretty_print_node_type(node, game.world_list)
        if description_line:
            print_function(f"  * {description_line}")

        for target_node, requirement in game.world_list.area_connections_from(node):
            print_function("  > {}".format(target_node.name))
            for level, text in pretty_print_requirement(requirement.simplify()):
                print_function("      {}{}".format("    " * level, text))
        print_function()


def write_human_readable_world_list(game: GameDescription, output: TextIO) -> None:
    def print_to_file(*args):
        output.write("\t".join(str(arg) for arg in args) + "\n")

    output.write("====================\nTemplates\n")
    for template_name, template in game.resource_database.requirement_template.items():
        output.write(f"\n* {template_name}:\n")
        for level, text in pretty_print_requirement(template):
            output.write("      {}{}\n".format("    " * level, text))

    output.write("\n====================\nDock Weaknesses\n")
    for dock_type in iterate_enum(DockType):
        output.write(f"\n> {dock_type}")
        for weakness in game.dock_weakness_database.get_by_type(dock_type):
            output.write(f"\n  * ({weakness.index}) {weakness.name}; Blast Shield? {weakness.is_blast_shield}\n")
            for level, text in pretty_print_requirement(weakness.requirement):
                output.write("      {}{}\n".format("    " * level, text))

    output.write("\n")
    for world in game.world_list.worlds:
        output.write("====================\n{}\n".format(world.name))
        for area in world.areas:
            output.write("----------------\n")
            pretty_print_area(game, area, print_function=print_to_file)



import dataclasses
from typing import Tuple

from randovania.bitpacking.bitpacking import BitPackDataClass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.game_description.echoes_game_specific import EchoesBeamConfiguration
from randovania.game_description.resources.resource_database import ResourceDatabase


@dataclasses.dataclass(frozen=True)
class BeamAmmoConfiguration(BitPackDataClass, JsonDataclass):
    item_index: int = dataclasses.field(metadata={"min": 0, "max": 108})
    ammo_a: int = dataclasses.field(metadata={"min": -1, "max": 108})
    ammo_b: int = dataclasses.field(metadata={"min": -1, "max": 108})
    uncharged_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 1})
    charged_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 5})
    combo_missile_cost: int = dataclasses.field(metadata={"min": 1, "max": 250, "if_different": 5})
    combo_ammo_cost: int = dataclasses.field(metadata={"min": 0, "max": 250, "if_different": 30})

    def create_game_specific(self, resource_database: ResourceDatabase) -> EchoesBeamConfiguration:
        return EchoesBeamConfiguration(
            item=resource_database.get_item(self.item_index),
            ammo_a=resource_database.get_item(self.ammo_a) if self.ammo_a >= 0 else None,
            ammo_b=resource_database.get_item(self.ammo_b) if self.ammo_b >= 0 else None,
            uncharged_cost=self.uncharged_cost,
            charged_cost=self.charged_cost,
            combo_missile_cost=self.combo_missile_cost,
            combo_ammo_cost=self.combo_ammo_cost,
        )


@dataclasses.dataclass(frozen=True)
class BeamConfiguration(BitPackDataClass, JsonDataclass):
    power: BeamAmmoConfiguration
    dark: BeamAmmoConfiguration
    light: BeamAmmoConfiguration
    annihilator: BeamAmmoConfiguration

    def create_game_specific(self, resource_database: ResourceDatabase) -> Tuple[EchoesBeamConfiguration, ...]:
        return (
            self.power.create_game_specific(resource_database),
            self.dark.create_game_specific(resource_database),
            self.light.create_game_specific(resource_database),
            self.annihilator.create_game_specific(resource_database),
        )



from typing import List, TypeVar, Callable, Dict, Tuple, Iterator

from randovania.game_description.area import Area
from randovania.game_description.dock import DockWeaknessDatabase, DockWeakness
from randovania.game_description.echoes_game_specific import EchoesBeamConfiguration, EchoesGameSpecific
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import Node, GenericNode, DockNode, PickupNode, TeleporterNode, EventNode, \
    TranslatorGateNode, LogbookNode, LoreType, PlayerShipNode
from randovania.game_description.requirements import ResourceRequirement, \
    RequirementOr, RequirementAnd, Requirement, RequirementTemplate
from randovania.game_description.resources.damage_resource_info import DamageResourceInfo
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo, ResourceGainTuple, ResourceGain
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.game_description.world import World
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame


def write_resource_requirement(requirement: ResourceRequirement) -> dict:
    return {
        "type": "resource",
        "data": {
            "type": requirement.resource.resource_type.value,
            "index": requirement.resource.index,
            "amount": requirement.amount,
            "negate": requirement.negate,
        }
    }


def write_requirement_and(requirement: RequirementAnd) -> dict:
    return {
        "type": "and",
        "data": [
            write_requirement(item)
            for item in requirement.items
        ]
    }


def write_requirement_or(requirement: RequirementOr) -> dict:
    return {
        "type": "or",
        "data": [
            write_requirement(item)
            for item in requirement.items
        ]
    }


def write_requirement_template(requirement: RequirementTemplate) -> dict:
    return {
        "type": "template",
        "data": requirement.template_name
    }


def write_requirement(requirement: Requirement) -> dict:
    if isinstance(requirement, ResourceRequirement):
        return write_resource_requirement(requirement)

    elif isinstance(requirement, RequirementOr):
        return write_requirement_or(requirement)

    elif isinstance(requirement, RequirementAnd):
        return write_requirement_and(requirement)

    elif isinstance(requirement, RequirementTemplate):
        return write_requirement_template(requirement)

    else:
        raise ValueError(f"Unknown requirement type: {type(requirement)}")


# Resource

def write_resource_gain(resource_gain: ResourceGain) -> list:
    def sorter(item: Tuple[ResourceInfo, int]):
        return item[0].resource_type, item[0].index, item[1]

    return [
        {
            "resource_type": resource.resource_type.value,
            "resource_index": resource.index,
            "amount": gain,
        }
        for resource, gain in sorted(resource_gain, key=sorter)
    ]


def write_simple_resource(resource: SimpleResourceInfo) -> dict:
    return {
        "index": resource.index,
        "long_name": resource.long_name,
        "short_name": resource.short_name,
    }


def write_item_resource(resource: ItemResourceInfo) -> dict:
    return {
        "index": resource.index,
        "long_name": resource.long_name,
        "short_name": resource.short_name,
        "max_capacity": resource.max_capacity,
        "extra": resource.extra,
    }


def write_trick_resource(resource: TrickResourceInfo) -> dict:
    return {
        "index": resource.index,
        "long_name": resource.long_name,
        "short_name": resource.short_name,
        "description": resource.description,
    }


def write_damage_resource(resource: DamageResourceInfo) -> dict:
    return {
        "index": resource.index,
        "long_name": resource.long_name,
        "short_name": resource.short_name,
        "reductions": [
            {
                "index": reduction.inventory_item.index,
                "multiplier": reduction.damage_multiplier
            }
            for reduction in resource.reductions
        ]
    }


X = TypeVar('X')


def write_array(array: List[X], writer: Callable[[X], dict]) -> list:
    return [
        writer(item)
        for item in array
    ]


def check_for_duplicated_index(array: List) -> Iterator[str]:
    indices_seen = set()
    for item in array:
        if item.index in indices_seen:
            yield f"Duplicated index {item.index} with {item.long_name}"
        else:
            indices_seen.add(item.index)


def write_resource_database(resource_database: ResourceDatabase):
    errors = []
    for array in (resource_database.item, resource_database.event, resource_database.trick, resource_database.damage,
                  resource_database.version, resource_database.misc):
        errors.extend(check_for_duplicated_index(array))

    if errors:
        raise ValueError("Errors in resource database: {}".format("\n".join(errors)))

    return {
        "items": write_array(resource_database.item, write_item_resource),
        "energy_tank_item_index": resource_database.energy_tank_item_index,
        "item_percentage_index": resource_database.item_percentage_index,
        "multiworld_magic_item_index": resource_database.multiworld_magic_item_index,
        "events": write_array(resource_database.event, write_simple_resource),
        "tricks": write_array(resource_database.trick, write_trick_resource),
        "damage": write_array(resource_database.damage, write_damage_resource),
        "versions": write_array(resource_database.version, write_simple_resource),
        "misc": write_array(resource_database.misc, write_simple_resource),
        "requirement_template": {
            name: write_requirement(requirement)
            for name, requirement in resource_database.requirement_template.items()
        }
    }


# Dock Weakness Database

def write_dock_weakness(dock_weakness: DockWeakness) -> dict:
    return {
        "index": dock_weakness.index,
        "name": dock_weakness.name,
        "is_blast_door": dock_weakness.is_blast_shield,
        "requirement": write_requirement(dock_weakness.requirement)
    }


def write_dock_weakness_database(database: DockWeaknessDatabase) -> dict:
    errors = []
    for array in (database.door, database.portal, database.morph_ball):
        errors.extend(check_for_duplicated_index(array))

    if errors:
        raise ValueError("Errors in dock weaknesses: {}".format("\n".join(errors)))

    return {
        "door": [
            write_dock_weakness(weakness)
            for weakness in database.door
        ],
        "portal": [
            write_dock_weakness(weakness)
            for weakness in database.portal
        ],
        "morph_ball": [
            write_dock_weakness(weakness)
            for weakness in database.morph_ball
        ],
    }


# Game Specific

def write_echoes_beam_configuration(beam_config: EchoesBeamConfiguration) -> dict:
    return {
        "item_index": beam_config.item.index,
        "ammo_a": beam_config.ammo_a.index if beam_config.ammo_a is not None else None,
        "ammo_b": beam_config.ammo_b.index if beam_config.ammo_b is not None else None,
        "uncharged_cost": beam_config.uncharged_cost,
        "charged_cost": beam_config.charged_cost,
        "combo_missile_cost": beam_config.combo_missile_cost,
        "combo_ammo_cost": beam_config.combo_ammo_cost,
    }


def write_game_specific(game_specific: EchoesGameSpecific) -> dict:
    return {
        "energy_per_tank": game_specific.energy_per_tank,
        "safe_zone_heal_per_second": game_specific.safe_zone_heal_per_second,
        "beam_configurations": [
            write_echoes_beam_configuration(beam)
            for beam in game_specific.beam_configurations
        ],
        "dangerous_energy_tank": game_specific.dangerous_energy_tank,
    }


# World/Area/Nodes

def write_node(node: Node) -> dict:
    """
    :param node:
    :return:
    """

    data = {
        "name": node.name,
        "heal": node.heal,
        "coordinates": {"x": node.location.x, "y": node.location.y, "z": node.location.z} if node.location else None,
    }

    if isinstance(node, GenericNode):
        data["node_type"] = "generic"

    elif isinstance(node, DockNode):
        data["node_type"] = "dock"
        data["dock_index"] = node.dock_index
        data["connected_area_asset_id"] = node.default_connection.area_asset_id
        data["connected_dock_index"] = node.default_connection.dock_index
        data["dock_type"] = node.default_dock_weakness.dock_type.value
        data["dock_weakness_index"] = node.default_dock_weakness.index

    elif isinstance(node, PickupNode):
        data["node_type"] = "pickup"
        data["pickup_index"] = node.pickup_index.index
        data["major_location"] = node.major_location

    elif isinstance(node, TeleporterNode):
        data["node_type"] = "teleporter"
        data["destination_world_asset_id"] = node.default_connection.world_asset_id
        data["destination_area_asset_id"] = node.default_connection.area_asset_id
        data["teleporter_instance_id"] = node.teleporter_instance_id
        data["scan_asset_id"] = node.scan_asset_id
        data["keep_name_when_vanilla"] = node.keep_name_when_vanilla
        data["editable"] = node.editable

    elif isinstance(node, EventNode):
        data["node_type"] = "event"
        data["event_index"] = node.resource().index
        if not node.name.startswith("Event -"):
            raise ValueError(f"'{node.name}' is an Event Node, but naming doesn't start with 'Event -'")

    elif isinstance(node, TranslatorGateNode):
        data["node_type"] = "translator_gate"
        data["gate_index"] = node.gate.index

    elif isinstance(node, LogbookNode):
        data["node_type"] = "logbook"
        data["string_asset_id"] = node.string_asset_id
        data["lore_type"] = node.lore_type.value

        if node.lore_type == LoreType.LUMINOTH_LORE:
            data["extra"] = node.required_translator.index

        elif node.lore_type in {LoreType.LUMINOTH_WARRIOR, LoreType.SKY_TEMPLE_KEY_HINT}:
            data["extra"] = node.hint_index
        else:
            data["extra"] = 0

    elif isinstance(node, PlayerShipNode):
        data["node_type"] = "player_ship"
        data["is_unlocked"] = write_requirement(node.is_unlocked)

    else:
        raise ValueError("Unknown node class: {}".format(node))

    if node.name.startswith("Event -") and data["node_type"] != "event":
        raise ValueError(f"'{node.name}' is not an Event Node, but naming suggests it is.")

    return data


def write_area(area: Area) -> dict:
    """
    :param area:
    :return:
    """
    errors = []

    nodes = []
    for node in area.nodes:
        try:
            data = write_node(node)
            data["connections"] = {
                target_node.name: write_requirement(area.connections[node][target_node])
                for target_node in area.nodes
                if target_node in area.connections[node]
            }
            nodes.append(data)
        except ValueError as e:
            errors.append(str(e))

    if errors:
        raise ValueError("Area {} nodes has the following errors:\n* {}".format(
            area.name, "\n* ".join(errors)))

    return {
        "name": area.name,
        "in_dark_aether": area.in_dark_aether,
        "asset_id": area.area_asset_id,
        "default_node_index": area.default_node_index,
        "valid_starting_location": area.valid_starting_location,
        "nodes": nodes
    }


def write_world(world: World) -> dict:
    errors = []
    areas = []
    for area in world.areas:
        try:
            areas.append(write_area(area))
        except ValueError as e:
            errors.append(str(e))

    if errors:
        raise ValueError("World {} has the following errors:\n> {}".format(
            world.name, "\n\n> ".join(errors)))

    return {
        "name": world.name,
        "dark_name": world.dark_name,
        "asset_id": world.world_asset_id,
        "areas": areas,
    }


def write_world_list(world_list: WorldList) -> list:
    errors = []

    worlds = []
    for world in world_list.worlds:
        try:
            worlds.append(write_world(world))
        except ValueError as e:
            errors.append(str(e))

    if errors:
        raise ValueError("\n\n".join(errors))

    return worlds


# Game Description

def write_initial_states(initial_states: Dict[str, ResourceGainTuple]) -> dict:
    return {
        name: write_resource_gain(initial_state)
        for name, initial_state in initial_states.items()
    }


def write_game_description(game: GameDescription) -> dict:
    return {
        "game": game.game.value,
        "resource_database": write_resource_database(game.resource_database),

        "game_specific": write_game_specific(game.game_specific) if game.game == RandovaniaGame.PRIME2 else {},
        "starting_location": game.starting_location.as_json,
        "initial_states": write_initial_states(game.initial_states),
        "victory_condition": write_requirement(game.victory_condition),

        "dock_weakness_database": write_dock_weakness_database(game.dock_weakness_database),
        "worlds": write_world_list(game.world_list),
    }



from enum import Enum
from random import Random
from typing import Dict, List

from randovania.game_description.node import PickupNode
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world_list import WorldList


class LocationHintCreator:
    world_list: WorldList
    joke_hints: List[str]
    index_to_node: Dict[PickupIndex, PickupNode]

    def __init__(self, world_list: WorldList, rng: Random, base_joke_hints: List[str]):
        self.world_list = world_list
        self.rng = rng
        self.base_joke_hints = base_joke_hints
        self.joke_hints = []

        self.index_to_node = {
            node.pickup_index: node
            for node in world_list.all_nodes
            if isinstance(node, PickupNode)
        }

    def index_node_name(self, pickup_index: PickupIndex, hide_area: bool) -> str:
        return self.node_name(self.index_to_node[pickup_index], hide_area)

    def node_name(self, pickup_node: PickupNode, hide_area: bool) -> str:
        if hide_area:
            return self.world_list.world_name_from_node(pickup_node, True)
        else:
            return self.world_list.area_name(self.world_list.nodes_to_area(pickup_node))

    def create_joke_hint(self) -> str:
        if not self.joke_hints:
            self.joke_hints = sorted(self.base_joke_hints)
            self.rng.shuffle(self.joke_hints)
        return self.joke_hints.pop()


class TextColor(Enum):
    GUARDIAN = "#FF3333"
    ITEM = "#FF6705B3"
    JOKE = "#45F731"
    LOCATION = "#FF3333"


def color_text(color: TextColor, text: str):
    return f"&push;&main-color={color.value};{text}&pop;"


def create_simple_logbook_hint(asset_id: int, hint: str) -> dict:
    return {
        "asset_id": asset_id,
        "strings": [hint, "", hint],
    }



from PySide2.QtCore import QSize, Qt
from PySide2.QtGui import QPixmap, QImage, QPainter


def paint_with_opacity(pixmap: QPixmap, opacity: float):
    transparent_image = QImage(QSize(36, 36), QImage.Format_ARGB32_Premultiplied)
    transparent_image.fill(Qt.transparent)
    painter = QPainter(transparent_image)
    painter.setOpacity(opacity)
    painter.drawPixmap(18 - pixmap.width() / 2,
                       18 - pixmap.height() / 2,
                       pixmap)
    painter.end()
    return QPixmap.fromImage(transparent_image)



import dataclasses

import pytest

from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo


def make(name: str) -> SimpleResourceInfo:
    return SimpleResourceInfo(ord(name[0]), name, name, ResourceType.ITEM)


@pytest.mark.parametrize("items", [
    {},
    {make("A"): 1},
    {make("A"): 1, make("B"): 1},
])
def test_assign_extra_initial_items_to_empty(empty_patches, items):
    # Run
    new_patches = empty_patches.assign_extra_initial_items(items)

    # Assert
    assert new_patches.starting_items == items


@pytest.mark.parametrize(["initial", "new_items", "expected"], [
    ({}, {}, {}),
    ({make("A"): 1}, {}, {make("A"): 1}),
    ({make("A"): 1}, {make("B"): 1}, {make("A"): 1, make("B"): 1}),
    ({make("A"): 1}, {make("A"): 1}, {make("A"): 2}),
])
def test_assign_extra_initial_items_merge(empty_patches, initial, new_items, expected):
    # Setup
    initial_patches = dataclasses.replace(empty_patches, starting_items=initial)

    # Run
    new_patches = initial_patches.assign_extra_initial_items(new_items)

    # Assert
    assert new_patches.starting_items == expected



import asyncio
import json
import os
from pathlib import Path
from typing import List, Optional, Iterator, Dict

import randovania
from randovania.games.game import RandovaniaGame
from randovania.layout.preset_migration import VersionedPreset


def read_preset_list() -> List[Path]:
    base_path = randovania.get_data_path().joinpath("presets")

    with base_path.joinpath("presets.json").open() as presets_file:
        preset_list = json.load(presets_file)["presets"]

    return [
        base_path.joinpath(preset["path"])
        for preset in preset_list
    ]


class PresetManager:
    included_presets: List[VersionedPreset]
    custom_presets: Dict[str, VersionedPreset]
    _data_dir: Optional[Path]

    def __init__(self, data_dir: Optional[Path]):
        self.included_presets = [VersionedPreset.from_file_sync(f) for f in read_preset_list()]

        self.custom_presets = {}
        if data_dir is not None:
            self._data_dir = data_dir.joinpath("presets")
        else:
            self._data_dir = None

    async def load_user_presets(self):
        all_files = self._data_dir.glob(f"*.{VersionedPreset.file_extension()}")
        user_presets = await asyncio.gather(*[VersionedPreset.from_file(f) for f in all_files])
        for preset in user_presets:
            if preset.name in self.custom_presets or self.included_preset_with_name(preset.name) is not None:
                continue
            self.custom_presets[preset.name] = preset

    @property
    def default_preset(self) -> VersionedPreset:
        return self.included_presets[0]

    def default_preset_for_game(self, game: RandovaniaGame) -> VersionedPreset:
        for preset in self.included_presets:
            if preset.game == game:
                return preset
        raise ValueError(f"{game} has no included preset")

    @property
    def all_presets(self) -> Iterator[VersionedPreset]:
        yield from self.included_presets
        yield from self.custom_presets.values()

    def add_new_preset(self, new_preset: VersionedPreset) -> bool:
        """
        Adds a new custom preset.
        :param: new_preset
        :return True, if there wasn't any preset with that name
        """
        if self.included_preset_with_name(new_preset.name) is not None:
            raise ValueError("A default preset with name '{}' already exists.".format(new_preset.name))

        existed_before = new_preset.name in self.custom_presets
        self.custom_presets[new_preset.name] = new_preset

        path = self._file_name_for_preset(new_preset)
        new_preset.save_to_file(path)
        return not existed_before

    def delete_preset(self, preset: VersionedPreset):
        del self.custom_presets[preset.name]
        os.remove(self._file_name_for_preset(preset))

    def included_preset_with_name(self, preset_name: str) -> Optional[VersionedPreset]:
        for preset in self.included_presets:
            if preset.name == preset_name:
                return preset

        return None

    def preset_for_name(self, preset_name: str) -> Optional[VersionedPreset]:
        preset = self.included_preset_with_name(preset_name)
        if preset is not None:
            return preset
        return self.custom_presets.get(preset_name)

    def _file_name_for_preset(self, preset: VersionedPreset) -> Path:
        return self._data_dir.joinpath("{}.{}".format(preset.slug_name, preset.file_extension()))



from randovania.dol_patching.assembler import ppc


def test_instruction_iter():
    assert list(ppc.Instruction(1234).bytes_for(0)) == [0x00, 0x00, 0x04, 0xD2]


def test_instruction_eq():
    assert ppc.Instruction(1234) == ppc.Instruction(1234)


def test_instruction_with_label():
    instruction = ppc.Instruction(1234)
    assert instruction.label is None

    with_label = instruction.with_label("foobar")
    assert instruction is with_label
    assert instruction.label == "foobar"


def test_stw():
    assert list(ppc.stw(ppc.r31, 0x1c, ppc.r1).bytes_for(0)) == [0x93, 0xe1, 0x00, 0x1c]


def test_stfs():
    assert list(ppc.stfs(ppc.f0, 0x1000, ppc.r2).bytes_for(0)) == [208, 2, 16, 0]


def test_lfs():
    assert list(ppc.lfs(ppc.f0, (0x8041a4a8 - 0x804223c0), ppc.r2).bytes_for(0)) == [0xc0, 0x02, 0x80, 0xe8]


def test_addi():
    assert list(ppc.addi(ppc.r3, ppc.r1, 0x8).bytes_for(0)) == [0x38, 0x61, 0x00, 0x08]


def test_or():
    assert list(ppc.or_(ppc.r31, ppc.r3, ppc.r3).bytes_for(0)) == [124, 127, 27, 120]


def test_lmw():
    assert list(ppc.lmw(ppc.r25, 0x774, ppc.r25).bytes_for(0)) == [187, 57, 7, 116]


def test_lwz():
    assert list(ppc.lwz(ppc.r10, 0x774, ppc.r25).bytes_for(0)) == [0x81, 0x59, 0x07, 0x74]


def test_lbz():
    assert list(ppc.lbz(ppc.r4, 0x2, ppc.r3).bytes_for(0)) == [0x88, 0x83, 0x00, 0x02]


def test_cmpwi_a():
    assert list(ppc.cmpwi(ppc.r30, 0).bytes_for(0)) == [0x2c, 0x1e, 0x00, 0x00]


def test_cmpwi_b():
    assert list(ppc.cmpwi(ppc.r4, -1).bytes_for(0)) == [0x2c, 0x04, 0xFF, 0xFF]


def test_b():
    assert list(ppc.b(0x80085760).bytes_for(0x80085760 + 0x9C)) == [0x4b, 0xff, 0xff, 0x64]


def test_bl():
    assert list(ppc.bl(0x80085760).bytes_for(0x80085760 + 0x9C)) == [0x4b, 0xff, 0xff, 0x65]


def test_beq():
    assert list(ppc.beq(0x80038094).bytes_for(0x80038034)) == [0x41, 0x82, 0x00, 0x60]


def test_bne():
    assert list(ppc.bne(0x80038094).bytes_for(0x80038034)) == [0x40, 0x82, 0x00, 0x60]


def test_blr():
    assert list(ppc.blr().bytes_for(0)) == [0x4e, 0x80, 0x00, 0x20]


def test_bctrl():
    assert list(ppc.bctrl().bytes_for(0)) == [0x4E, 0x80, 0x04, 0x21]


def test_li():
    assert list(ppc.li(ppc.r5, 9999).bytes_for(0)) == [56, 160, 39, 15]


def test_stwu():
    assert list(ppc.stwu(ppc.r1, -0x2C, ppc.r1).bytes_for(0)) == [0x94, 0x21, 0xff, 0xd4]


def test_stmw():
    assert list(ppc.stmw(ppc.r25, -0x2C, ppc.r1).bytes_for(0)) == [191, 33, 255, 212]


def test_sync():
    assert list(ppc.sync().bytes_for(0)) == [0x7c, 0x00, 0x04, 0xac]


def test_isync():
    assert list(ppc.isync().bytes_for(0)) == [0x4c, 0x00, 0x01, 0x2c]


def test_dcbi():
    assert list(ppc.dcbi(1, 2).bytes_for(0)) == [0x7c, 0x01, 0x13, 0xac]


def test_mfspr_LR():
    assert list(ppc.mfspr(ppc.r0, ppc.LR).bytes_for(0)) == [0x7c, 0x08, 0x02, 0xa6]


def test_mtctr():
    assert list(ppc.mtctr(ppc.r12).bytes_for(0)) == [0x7d, 0x89, 0x03, 0xa6]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'corruption_layout_editor.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_CorruptionLayoutEditor(object):
    def setupUi(self, CorruptionLayoutEditor):
        if not CorruptionLayoutEditor.objectName():
            CorruptionLayoutEditor.setObjectName(u"CorruptionLayoutEditor")
        CorruptionLayoutEditor.resize(1025, 508)
        self.actionAction = QAction(CorruptionLayoutEditor)
        self.actionAction.setObjectName(u"actionAction")
        self.menu_reset_action = QAction(CorruptionLayoutEditor)
        self.menu_reset_action.setObjectName(u"menu_reset_action")
        self.central_widget = QWidget(CorruptionLayoutEditor)
        self.central_widget.setObjectName(u"central_widget")
        self.central_widget.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout = QGridLayout(self.central_widget)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(4, -1, 4, -1)
        self.scroll_area = QScrollArea(self.central_widget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 1015, 439))
        self.scroll_area_layout = QHBoxLayout(self.scroll_area_contents)
        self.scroll_area_layout.setSpacing(6)
        self.scroll_area_layout.setContentsMargins(11, 11, 11, 11)
        self.scroll_area_layout.setObjectName(u"scroll_area_layout")
        self.scroll_area_layout.setContentsMargins(2, 2, 2, 2)
        self.scroll_area.setWidget(self.scroll_area_contents)

        self.gridLayout.addWidget(self.scroll_area, 0, 0, 1, 2)

        self.layout_label = QLabel(self.central_widget)
        self.layout_label.setObjectName(u"layout_label")

        self.gridLayout.addWidget(self.layout_label, 1, 0, 1, 1)

        self.layout_edit = QLineEdit(self.central_widget)
        self.layout_edit.setObjectName(u"layout_edit")
        self.layout_edit.setReadOnly(True)

        self.gridLayout.addWidget(self.layout_edit, 1, 1, 1, 1)

        CorruptionLayoutEditor.setCentralWidget(self.central_widget)
        self.menu_bar = QMenuBar(CorruptionLayoutEditor)
        self.menu_bar.setObjectName(u"menu_bar")
        self.menu_bar.setGeometry(QRect(0, 0, 1025, 21))
        CorruptionLayoutEditor.setMenuBar(self.menu_bar)

        self.retranslateUi(CorruptionLayoutEditor)

        QMetaObject.connectSlotsByName(CorruptionLayoutEditor)
    # setupUi

    def retranslateUi(self, CorruptionLayoutEditor):
        CorruptionLayoutEditor.setWindowTitle(QCoreApplication.translate("CorruptionLayoutEditor", u"Tracker", None))
        self.actionAction.setText(QCoreApplication.translate("CorruptionLayoutEditor", u"Action", None))
        self.menu_reset_action.setText(QCoreApplication.translate("CorruptionLayoutEditor", u"Reset", None))
        self.layout_label.setText(QCoreApplication.translate("CorruptionLayoutEditor", u"Layout string:", None))
    # retranslateUi




import json
from typing import Dict

from randovania import get_data_path
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.layout.translator_configuration import LayoutTranslatorRequirement


def _raw_translator_configurations(data: dict) -> Dict[TranslatorGate, LayoutTranslatorRequirement]:
    return {
        TranslatorGate(int(key)): LayoutTranslatorRequirement(item)
        for key, item in data["translator_requirement"].items()
    }


def get_vanilla_actual_translator_configurations() -> Dict[TranslatorGate, LayoutTranslatorRequirement]:
    with get_data_path().joinpath("item_database", "prime2", "default_state",
                                  "translator_vanilla_actual.json").open() as open_file:
        data = json.load(open_file)

    return _raw_translator_configurations(data)


def get_vanilla_colors_translator_configurations() -> Dict[TranslatorGate, LayoutTranslatorRequirement]:
    with get_data_path().joinpath("item_database", "prime2", "default_state",
                                  "translator_vanilla_colors.json").open() as open_file:
        data = json.load(open_file)

    return _raw_translator_configurations(data)



from typing import Iterable, Dict, NamedTuple, Optional

from PySide2 import QtWidgets
from PySide2.QtCore import Qt

from randovania.game_description.item.ammo import Ammo
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.ammo_state import AmmoState
from randovania.layout.preset import Preset


class AmmoPickupWidgets(NamedTuple):
    pickup_spinbox: QtWidgets.QSpinBox
    expected_count: QtWidgets.QLabel
    pickup_box: QtWidgets.QGroupBox
    require_major_item_check: Optional[QtWidgets.QCheckBox]


class SplitAmmoWidget(QtWidgets.QCheckBox):
    def __init__(self, parent: QtWidgets.QWidget, editor: PresetEditor,
                 unified_ammo: Ammo, split_ammo: Iterable[Ammo]):
        super().__init__(parent)
        self._editor = editor
        self.unified_ammo = unified_ammo
        self.split_ammo = list(split_ammo)
        self.setTristate(True)
        self._last_check_state = (True, False)
        self.clicked.connect(self.change_split)

    def on_preset_changed(self, preset: Preset, ammo_pickup_widgets: Dict[Ammo, AmmoPickupWidgets]):
        ammo_configuration = preset.configuration.ammo_configuration

        has_unified = ammo_configuration.items_state[self.unified_ammo].pickup_count > 0
        has_split = any(ammo_configuration.items_state[item].pickup_count > 0
                        for item in self.split_ammo)

        if not has_split and not has_unified:
            has_split, has_unified = self._last_check_state
        else:
            self._last_check_state = has_split, has_unified

        if has_unified:
            new_state = Qt.PartiallyChecked if has_split else Qt.Unchecked
        else:
            new_state = Qt.Checked
        self.setCheckState(new_state)

        ammo_pickup_widgets[self.unified_ammo].pickup_box.setVisible(has_unified)
        for item in self.split_ammo:
            ammo_pickup_widgets[item].pickup_box.setVisible(has_split)

    def change_split(self, has_split: bool):
        with self._editor as editor:
            ammo_configuration = editor.ammo_configuration

            current_total = sum(
                ammo_configuration.items_state[ammo].pickup_count
                for ammo in (self.unified_ammo, *self.split_ammo)
            )
            if has_split:
                split_state = AmmoState(pickup_count=current_total // len(self.split_ammo))
                unified_state = AmmoState()
            else:
                split_state = AmmoState()
                unified_state = AmmoState(pickup_count=current_total)

            new_states = {self.unified_ammo: unified_state}
            for ammo in self.split_ammo:
                new_states[ammo] = split_state

            editor.ammo_configuration = ammo_configuration.replace_states(new_states)
            self._last_check_state = (has_split, not has_split)



import collections
import dataclasses
import re
from typing import Dict, List, DefaultDict

from randovania.game_description import data_reader
from randovania.game_description.area import Area
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import PickupAssignment, PickupTarget
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint
from randovania.game_description.node import PickupNode, TeleporterNode
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import find_resource_info_with_long_name
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world_list import WorldList
from randovania.games.prime import default_data
from randovania.generator import base_patches_factory
from randovania.generator.item_pool import pool_creator, PoolResults
from randovania.layout.echoes_configuration import EchoesConfiguration

_ETM_NAME = "Energy Transfer Module"


def _pickup_assignment_to_item_locations(world_list: WorldList,
                                         pickup_assignment: PickupAssignment,
                                         num_players: int,
                                         ) -> Dict[str, Dict[str, str]]:
    items_locations: DefaultDict[str, Dict[str, str]] = collections.defaultdict(dict)

    for world, area, node in world_list.all_worlds_areas_nodes:
        if not node.is_resource_node or not isinstance(node, PickupNode):
            continue

        if node.pickup_index in pickup_assignment:
            target = pickup_assignment[node.pickup_index]
            if num_players > 1:
                item_name = f"{target.pickup.name} for Player {target.player}"
            else:
                item_name = f"{target.pickup.name}"
        else:
            item_name = _ETM_NAME

        world_name = world.dark_name if area.in_dark_aether else world.name
        items_locations[world_name][world_list.node_name(node)] = item_name

    return {
        world: {
            area: item
            for area, item in sorted(items_locations[world].items())
        }
        for world in sorted(items_locations.keys())
    }


def _node_mapping_to_elevator_connection(world_list: WorldList,
                                         elevators: Dict[str, str],
                                         ) -> Dict[int, AreaLocation]:
    result = {}
    for source_name, target_node in elevators.items():
        source_node = world_list.node_from_name(source_name)
        assert isinstance(source_node, TeleporterNode)

        target_node = world_list.node_from_name(target_node)

        result[source_node.teleporter_instance_id] = AreaLocation(
            world_list.nodes_to_world(target_node).world_asset_id,
            world_list.nodes_to_area(target_node).area_asset_id
        )

    return result


def _find_area_with_teleporter(world_list: WorldList, teleporter_id: int) -> Area:
    for _, area, node in world_list.all_worlds_areas_nodes:
        if isinstance(node, TeleporterNode):
            if node.teleporter_instance_id == teleporter_id:
                return area
    raise ValueError("Unknown teleporter_id: {}".format(teleporter_id))


def _name_for_gate(gate: TranslatorGate) -> str:
    for items in default_data.decode_randomizer_data()["TranslatorLocationData"]:
        if items["Index"] == gate.index:
            return items["Name"]
    raise ValueError("Unknown gate: {}".format(gate))


def _find_gate_with_name(gate_name: str) -> TranslatorGate:
    for items in default_data.decode_randomizer_data()["TranslatorLocationData"]:
        if items["Name"] == gate_name:
            return TranslatorGate(items["Index"])
    raise ValueError("Unknown gate name: {}".format(gate_name))


def serialize_single(player_index: int, num_players: int, patches: GamePatches, game_data: dict) -> dict:
    """
    Encodes a given GamePatches into a JSON-serializable dict.
    :param player_index:
    :param num_players:
    :param patches:
    :param game_data:
    :return:
    """
    game = data_reader.decode_data(game_data)
    world_list = game.world_list

    result = {
        "starting_location": world_list.area_name(world_list.area_by_area_location(patches.starting_location),
                                                  separator="/"),
        "starting_items": {
            resource_info.long_name: quantity
            for resource_info, quantity in patches.starting_items.items()
        },
        "elevators": {
            world_list.area_name(_find_area_with_teleporter(world_list, teleporter_id), "/"):
                world_list.area_name(world_list.area_by_area_location(connection), "/")
            for teleporter_id, connection in patches.elevator_connection.items()
        },
        "translators": {
            _name_for_gate(gate): requirement.long_name
            for gate, requirement in patches.translator_gates.items()
        },
        "locations": {
            key: value
            for key, value in _pickup_assignment_to_item_locations(world_list,
                                                                   patches.pickup_assignment,
                                                                   num_players).items()
        },
        "hints": {
            str(asset.asset_id): hint.as_json
            for asset, hint in patches.hints.items()
        }
    }
    return result


def _area_name_to_area_location(world_list: WorldList, area_name: str) -> AreaLocation:
    world_name, area_name = re.match("([^/]+)/([^/]+)", area_name).group(1, 2)
    starting_world = world_list.world_with_name(world_name)
    starting_area = starting_world.area_by_name(area_name)
    return AreaLocation(starting_world.world_asset_id, starting_area.area_asset_id)


def _find_pickup_with_name(item_pool: List[PickupEntry], pickup_name: str) -> PickupEntry:
    for pickup in item_pool:
        if pickup.name == pickup_name:
            return pickup

    names = set(pickup.name for pickup in item_pool)
    raise ValueError(f"Unable to find a pickup with name {pickup_name}. Possible values: {sorted(names)}")


def decode_single(player_index: int, all_pools: Dict[int, PoolResults], game: GameDescription,
                  game_modifications: dict, configuration: EchoesConfiguration) -> GamePatches:
    """
    Decodes a dict created by `serialize` back into a GamePatches.
    :param player_index:
    :param all_pools:
    :param game:
    :param game_modifications:
    :param configuration:
    :return:
    """
    game_specific = base_patches_factory.create_game_specific(configuration, game)
    world_list = game.world_list

    initial_pickup_assignment = all_pools[player_index].assignment

    # Starting Location
    starting_location = _area_name_to_area_location(world_list, game_modifications["starting_location"])

    # Initial items
    starting_items = {
        find_resource_info_with_long_name(game.resource_database.item, resource_name): quantity
        for resource_name, quantity in game_modifications["starting_items"].items()
    }

    # Elevators
    elevator_connection = {}
    for source_name, target_name in game_modifications["elevators"].items():
        source_area = _area_name_to_area_location(world_list, source_name)
        target_area = _area_name_to_area_location(world_list, target_name)

        potential_source_nodes = [
            node
            for node in world_list.area_by_area_location(source_area).nodes
            if isinstance(node, TeleporterNode)
        ]
        assert len(potential_source_nodes) == 1
        source_node = potential_source_nodes[0]
        elevator_connection[source_node.teleporter_instance_id] = target_area

    # Translator Gates
    translator_gates = {
        _find_gate_with_name(gate_name): find_resource_info_with_long_name(game.resource_database.item, resource_name)
        for gate_name, resource_name in game_modifications["translators"].items()
    }

    # Pickups
    target_name_re = re.compile(r"(.*) for Player (\d+)")

    pickup_assignment: PickupAssignment = {}
    for world_name, world_data in game_modifications["locations"].items():
        for area_node_name, target_name in world_data.items():
            if target_name == _ETM_NAME:
                continue

            pickup_name_match = target_name_re.match(target_name)
            if pickup_name_match is not None:
                pickup_name = pickup_name_match.group(1)
                target_player = int(pickup_name_match.group(2))
            else:
                pickup_name = target_name
                target_player = 0

            node = world_list.node_from_name(f"{world_name}/{area_node_name}")
            assert isinstance(node, PickupNode)
            if node.pickup_index in initial_pickup_assignment:
                pickup = initial_pickup_assignment[node.pickup_index]
                if (pickup_name, target_player) != (pickup.name, player_index):
                    raise ValueError(f"{area_node_name} should be vanilla based on configuration")

            elif pickup_name != _ETM_NAME:
                configuration_item_pool = all_pools[target_player].pickups
                pickup = _find_pickup_with_name(configuration_item_pool, pickup_name)
                configuration_item_pool.remove(pickup)
            else:
                pickup = None

            if pickup is not None:
                pickup_assignment[node.pickup_index] = PickupTarget(pickup, target_player)

    # Hints
    hints = {}
    for asset_id, hint in game_modifications["hints"].items():
        hints[LogbookAsset(int(asset_id))] = Hint.from_json(hint)

    return GamePatches(
        player_index=player_index,
        pickup_assignment=pickup_assignment,  # PickupAssignment
        elevator_connection=elevator_connection,  # Dict[int, AreaLocation]
        dock_connection={},  # Dict[Tuple[int, int], DockConnection]
        dock_weakness={},  # Dict[Tuple[int, int], DockWeakness]
        translator_gates=translator_gates,
        starting_items=starting_items,  # ResourceGainTuple
        starting_location=starting_location,  # AreaLocation
        hints=hints,
        game_specific=game_specific,
    )


def decode(game_modifications: List[dict],
           layout_configurations: Dict[int, EchoesConfiguration],
           ) -> Dict[int, GamePatches]:

    all_games = {index: data_reader.decode_data(configuration.game_data)
                 for index, configuration in layout_configurations.items()}
    all_pools = {index: pool_creator.calculate_pool_results(configuration, all_games[index].resource_database)
                 for index, configuration in layout_configurations.items()}
    return {
        index: decode_single(index, all_pools, all_games[index], modifications, layout_configurations[index])
        for index, modifications in enumerate(game_modifications)
    }


def serialize(all_patches: Dict[int, GamePatches], all_game_data: Dict[int, dict]) -> List[dict]:
    return [
        serialize_single(index, len(all_patches), patches, all_game_data[index])
        for index, patches in all_patches.items()
    ]



import asyncio
import os
import subprocess
from typing import Union

import pytest
from PySide2.QtWidgets import QDialog
from mock import AsyncMock, MagicMock, patch, ANY

from randovania.gui.main_window import MainWindow
from randovania.interface_common.options import Options
from randovania.interface_common.preset_manager import PresetManager
from randovania.network_client.network_client import ConnectionState


def create_window(options: Union[Options, MagicMock],
                  preset_manager: PresetManager) -> MainWindow:
    return MainWindow(options, preset_manager, MagicMock(), False)


@pytest.fixture(name="default_main_window")
def _default_main_window(skip_qtbot, preset_manager) -> MainWindow:
    return create_window(Options(MagicMock()), preset_manager)


def test_drop_random_event(default_main_window: MainWindow,
                           ):
    # Creating a window should not fail
    pass


@pytest.mark.asyncio
@patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
@patch("randovania.gui.main_window.GameSessionBrowserDialog", autospec=True)
@patch("randovania.gui.main_window.GameSessionWindow.create_and_update", new_callable=AsyncMock)
async def test_browse_for_game_session(mock_game_session_window: AsyncMock,
                                       mock_game_session_browser: MagicMock,
                                       mock_execute_dialog: AsyncMock,
                                       skip_qtbot, default_main_window, mocker):
    # Setup
    mock_get_game_connection = mocker.patch("randovania.gui.lib.common_qt_lib.get_game_connection", autospec=True)
    mocker.patch("randovania.gui.main_window.MainWindow._ensure_logged_in", new_callable=AsyncMock,
                 return_value=True)
    mock_execute_dialog.return_value = mock_game_session_browser.return_value.Accepted
    mock_game_session_browser.return_value.refresh = AsyncMock()

    # Run
    await default_main_window._browse_for_game_session()

    # Assert
    mock_game_session_browser.assert_called_once_with(default_main_window.network_client)
    mock_game_session_browser.return_value.refresh.assert_awaited_once_with()
    mock_execute_dialog.assert_awaited_once_with(mock_game_session_browser.return_value)
    mock_game_session_window.assert_awaited_once_with(
        default_main_window.network_client,
        mock_get_game_connection.return_value,
        default_main_window.preset_manager,
        default_main_window,
        default_main_window._options,
    )
    mock_game_session_window.return_value.show.assert_called_once_with()


@pytest.mark.asyncio
@patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
@patch("randovania.gui.main_window.GameSessionWindow.create_and_update", new_callable=AsyncMock)
async def test_host_game_session(mock_game_session_window: MagicMock,
                                 mock_execute_dialog: AsyncMock,
                                 skip_qtbot, default_main_window, mocker):
    # Setup
    mock_get_game_connection = mocker.patch("randovania.gui.lib.common_qt_lib.get_game_connection", autospec=True)
    mocker.patch("randovania.gui.main_window.MainWindow._ensure_logged_in", new_callable=AsyncMock,
                 return_value=True)
    mock_execute_dialog.return_value = QDialog.Accepted
    default_main_window.network_client.create_new_session = AsyncMock()

    # Run
    await default_main_window._host_game_session()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    default_main_window.network_client.create_new_session.assert_awaited_once_with("")
    mock_game_session_window.assert_awaited_once_with(
        default_main_window.network_client,
        mock_get_game_connection.return_value,
        default_main_window.preset_manager,
        default_main_window,
        default_main_window._options,
    )
    mock_game_session_window.return_value.show.assert_called_once_with()


@pytest.mark.asyncio
async def test_ensure_logged_in(default_main_window, mocker):
    # Setup
    mock_message_box = mocker.patch("PySide2.QtWidgets.QMessageBox")

    async def true(): return True

    connect_task = asyncio.create_task(true())

    network_client = default_main_window.network_client
    network_client.connect_to_server.return_value = connect_task
    network_client.connection_state = ConnectionState.Disconnected
    network_client.current_user = MagicMock()

    # Run
    result = await default_main_window._ensure_logged_in()

    # Assert
    mock_message_box.assert_called_once()
    assert result


@pytest.mark.asyncio
async def test_browse_racetime(default_main_window, mocker):
    mock_new_dialog = mocker.patch("randovania.gui.dialog.racetime_browser_dialog.RacetimeBrowserDialog")
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=QDialog.Accepted)
    dialog = mock_new_dialog.return_value
    dialog.refresh = AsyncMock(return_value=True)
    default_main_window.generate_seed_from_permalink = AsyncMock()

    # Run
    await default_main_window._browse_racetime()

    # Assert
    mock_new_dialog.assert_called_once_with()
    dialog.refresh.assert_awaited_once_with()
    mock_execute_dialog.assert_awaited_once_with(dialog)
    default_main_window.generate_seed_from_permalink.assert_awaited_once_with(dialog.permalink)


@pytest.mark.asyncio
async def test_generate_seed_from_permalink(default_main_window, mocker):
    permalink = MagicMock()
    mock_generate_layout: MagicMock = mocker.patch("randovania.interface_common.simplified_patcher.generate_layout",
                                                   autospec=True)
    default_main_window.open_game_details = MagicMock()

    # Run
    await default_main_window.generate_seed_from_permalink(permalink)

    # Assert
    mock_generate_layout.assert_called_once_with(progress_update=ANY,
                                                 permalink=permalink,
                                                 options=default_main_window._options)
    default_main_window.open_game_details.assert_called_once_with(mock_generate_layout.return_value)


@pytest.mark.parametrize("os_type", ["Windows", "Darwin", "Linux"])
@pytest.mark.parametrize("throw_exception", [True, False])
def test_on_menu_action_previously_generated_games(default_main_window, mocker, os_type, throw_exception, monkeypatch):
    mock_start_file = MagicMock()
    mock_subprocess_run = MagicMock()
    monkeypatch.setattr(os, "startfile", mock_start_file, raising=False)
    monkeypatch.setattr(subprocess, "run", mock_subprocess_run, raising=False)
    mocker.patch("platform.system", return_value=os_type)
    mock_message_box = mocker.patch("PySide2.QtWidgets.QMessageBox")

    # Run
    if throw_exception:
        if os_type == "Windows":
            mock_start_file.side_effect = OSError()
        else:
            mock_subprocess_run.side_effect = OSError()

    default_main_window._on_menu_action_previously_generated_games()

    # Assert
    if throw_exception:
        mock_message_box.return_value.show.assert_called_once()
    else:
        if os_type == "Windows":
            mock_start_file.assert_called_once()
            mock_message_box.return_value.show.assert_not_called()
        else:
            mock_subprocess_run.assert_called_once()
            mock_message_box.return_value.show.assert_not_called()




@pytest.mark.asyncio
async def test_on_menu_action_map_tracker(default_main_window, mocker):
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=QDialog.Accepted)
    default_main_window.open_map_tracker = MagicMock()
    preset = next(default_main_window.preset_manager.all_presets)

    # Run
    await default_main_window._on_menu_action_map_tracker()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    default_main_window.open_map_tracker.assert_called_once_with(preset.get_preset().configuration)



from typing import Tuple

from randovania.dol_patching.assembler import ppc


class CompositeInstruction(ppc.BaseInstruction):
    def __init__(self, instructions: Tuple[ppc.BaseInstruction, ...]):
        super().__init__()
        self.instructions = instructions

    def bytes_for(self, address: int):
        for instruction in self.instructions:
            yield from instruction.bytes_for(address)
            address += instruction.byte_count

    def __eq__(self, other):
        return isinstance(other, CompositeInstruction) and other.instructions == self.instructions

    @property
    def byte_count(self):
        return sum(instruction.byte_count for instruction in self.instructions)


class CurrentAddressInstruction(ppc.BaseInstruction):
    def __init__(self, output_register: ppc.GeneralRegister, offset: int):
        super().__init__()
        self.output_register = output_register
        self.offset = offset

    def bytes_for(self, address: int):
        return load_unsigned_32bit(self.output_register, address + self.offset).bytes_for(address)

    def __eq__(self, other):
        return (isinstance(other, CurrentAddressInstruction) and
                (other.output_register, other.offset) == (self.output_register, self.offset))

    @property
    def byte_count(self):
        return 8


def load_unsigned_32bit(output_register: ppc.GeneralRegister, value: int) -> CompositeInstruction:
    return CompositeInstruction((
        ppc.lis(output_register, value >> 16),
        ppc.ori(output_register, output_register, value & 0xFFFF),
    ))


def load_current_address(output_register: ppc.GeneralRegister, instruction_offset: int = 0):
    return CurrentAddressInstruction(output_register, instruction_offset * 4)



from typing import Optional, List

from PySide2.QtWidgets import QWidget, QMainWindow

from randovania.games.game import RandovaniaGame
from randovania.gui.lib.close_event_widget import CloseEventWidget
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.layout_description import LayoutDescription


class WindowManager(QMainWindow):
    tracked_windows: List[CloseEventWidget]

    def __init__(self):
        super().__init__()
        self.tracked_windows = []

    @property
    def preset_manager(self) -> PresetManager:
        raise NotImplemented()

    def open_map_tracker(self, configuration: EchoesConfiguration):
        raise NotImplemented()

    def open_data_visualizer_at(self, world_name: Optional[str], area_name: Optional[str],
                                game: RandovaniaGame = RandovaniaGame.PRIME2):
        raise NotImplemented()

    def open_game_details(self, layout: LayoutDescription):
        raise NotImplemented()

    @property
    def main_window(self) -> QMainWindow:
        raise NotImplemented()

    @property
    def is_preview_mode(self) -> bool:
        raise NotImplemented()

    def track_window(self, window: CloseEventWidget):
        def remove_window():
            self.tracked_windows.remove(window)

        window.CloseEvent.connect(remove_window)
        self.tracked_windows.append(window)



from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class LayoutElevators(BitPackEnum, Enum):
    VANILLA = "vanilla"
    TWO_WAY_RANDOMIZED = "randomized"
    TWO_WAY_UNCHECKED = "two-way-unchecked"
    ONE_WAY_ELEVATOR = "one-way-elevator"
    ONE_WAY_ANYTHING = "one-way-anything"

    @classmethod
    def default(cls) -> "LayoutElevators":
        return cls.VANILLA



import multiprocessing
from concurrent.futures.process import ProcessPoolExecutor
from multiprocessing.connection import Connection
from typing import Callable

from randovania.generator import generator
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.resolver import debug


def _generate_layout_worker(output_pipe: Connection,
                            permalink: Permalink,
                            validate_after_generation: bool,
                            timeout_during_generation: bool,
                            debug_level: int):
    def status_update(message: str):
        output_pipe.send(message)
        if output_pipe.poll():
            raise RuntimeError(output_pipe.recv())

    debug.set_level(debug_level)
    extra_args = {}
    if not timeout_during_generation:
        extra_args["timeout"] = None

    return generator.generate_description(permalink,
                                          status_update=status_update,
                                          validate_after_generation=validate_after_generation,
                                          **extra_args)


def generate_layout(permalink: Permalink,
                    status_update: Callable[[str], None],
                    validate_after_generation: bool,
                    timeout_during_generation: bool,
                    ) -> LayoutDescription:
    receiving_pipe, output_pipe = multiprocessing.Pipe(True)

    debug_level = debug.debug_level()
    if not permalink.spoiler:
        debug_level = 0

    def on_done(_):
        output_pipe.send(None)

    with ProcessPoolExecutor(max_workers=1) as executor:
        future = executor.submit(_generate_layout_worker, output_pipe, permalink, validate_after_generation,
                                 timeout_during_generation, debug_level)

        future.add_done_callback(on_done)

        while not future.done():
            message = receiving_pipe.recv()
            if message is not None:
                try:
                    status_update(message)
                except Exception:
                    receiving_pipe.send("close")
                    raise

        return future.result()



from pathlib import Path
from unittest.mock import MagicMock, patch

from PySide2 import QtCore

from randovania.gui.dialog.game_input_dialog import GameInputDialog


@patch("randovania.gui.lib.common_qt_lib.prompt_user_for_output_iso", autospec=True)
def test_on_output_file_button_exists(mock_prompt_user_for_output_iso: MagicMock,
                                      skip_qtbot,
                                      tmpdir):
    # Setup
    temp_path = Path(tmpdir)
    options = MagicMock()
    options.output_directory = None
    window = GameInputDialog(options, "random game.iso", True)
    mock_prompt_user_for_output_iso.return_value = temp_path.joinpath("foo", "game.iso")

    # Run
    skip_qtbot.mouseClick(window.output_file_button, QtCore.Qt.LeftButton)

    # Assert
    mock_prompt_user_for_output_iso.assert_called_once_with(window,
                                                            "random game.iso")
    assert window.output_file_edit.text() == str(temp_path.joinpath("foo", "game.iso"))
    assert temp_path.joinpath("foo").is_dir()


@patch("randovania.gui.lib.common_qt_lib.prompt_user_for_output_iso", autospec=True)
def test_on_output_file_button_cancel(mock_prompt_user_for_output_iso: MagicMock,
                                      skip_qtbot,
                                      tmpdir):
    # Setup
    options = MagicMock()
    options.output_directory = None
    window = GameInputDialog(options, "random game.iso", True)
    mock_prompt_user_for_output_iso.return_value = None

    # Run
    skip_qtbot.mouseClick(window.output_file_button, QtCore.Qt.LeftButton)

    # Assert
    mock_prompt_user_for_output_iso.assert_called_once_with(window,
                                                            "random game.iso")
    assert window.output_file_edit.text() == ""



from PySide2.QtGui import QPixmap, QMouseEvent
from PySide2.QtWidgets import QLabel, QWidget


class ClickableLabel(QLabel):
    is_checked: bool

    def __init__(self, parent: QWidget, transparent_pixmap: QPixmap, opaque_pixmap: QPixmap):
        super().__init__(parent)
        self.transparent_pixmap = transparent_pixmap
        self.opaque_pixmap = opaque_pixmap
        self._ignoring_mouse_events = False

    def set_checked(self, is_opaque: bool):
        self.setPixmap(self.opaque_pixmap if is_opaque else self.transparent_pixmap)
        self.is_checked = is_opaque

    def set_ignore_mouse_events(self, enabled: bool):
        self._ignoring_mouse_events = enabled

    def mouseReleaseEvent(self, event_data: QMouseEvent):
        if self._ignoring_mouse_events:
            return
        if 0 <= event_data.x() <= self.width() and 0 <= event_data.y() <= self.height():
            event_data.accept()
            self.set_checked(not self.is_checked)
